#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from copy import deepcopy

import pytest

from pyxel.detectors import Characteristics


@pytest.mark.parametrize(
    "quantum_efficiency, charge_to_volt_conversion, pre_amplification, "
    "full_well_capacity, adc_voltage_range, adc_bit_resolution",
    [
        (0, 0, 0, 0, None, None),
        (1.0, 1.0, 100.0, 10_000_000, None, None),
        (1.0, 1.0, 100.0, 10_000_000, [0.0, 3.0], 16),
    ],
)
def test_create_valid_characteristics(
    quantum_efficiency,
    charge_to_volt_conversion,
    pre_amplification,
    full_well_capacity,
    adc_voltage_range,
    adc_bit_resolution,
):
    """Test when creating a valid `Characteristics` object."""
    _ = Characteristics(
        quantum_efficiency=quantum_efficiency,
        charge_to_volt_conversion=charge_to_volt_conversion,
        pre_amplification=pre_amplification,
        full_well_capacity=full_well_capacity,
        adc_voltage_range=adc_voltage_range,
        adc_bit_resolution=adc_bit_resolution,
    )


@pytest.mark.parametrize(
    "other_obj, is_equal",
    [
        pytest.param(None, False, id="None"),
        pytest.param(Characteristics(), False, id="Empty 'Characteristics'"),
        pytest.param(
            Characteristics(quantum_efficiency=0.1), False, id="Only one parameter"
        ),
        pytest.param(
            Characteristics(
                quantum_efficiency=0.1,
                charge_to_volt_conversion=0.2,
                pre_amplification=4.4,
            ),
            False,
            id="Wrong type",
        ),
        pytest.param(
            Characteristics(
                quantum_efficiency=0.1,
                charge_to_volt_conversion=0.2,
                pre_amplification=4.4,
                full_well_capacity=10,
            ),
            False,
            id="Missing some parameters",
        ),
        pytest.param(
            Characteristics(
                quantum_efficiency=0.1,
                charge_to_volt_conversion=0.2,
                pre_amplification=4.4,
                full_well_capacity=10,
                adc_voltage_range=(0.0, 10.0),
                adc_bit_resolution=16,
            ),
            True,
            id="Same parameters, same class",
        ),
    ],
)
def test_is_equal(other_obj, is_equal):
    """Test equality statement for `Characteristics`."""
    obj = Characteristics(
        quantum_efficiency=0.1,
        charge_to_volt_conversion=0.2,
        pre_amplification=4.4,
        full_well_capacity=10,
        adc_voltage_range=(0.0, 10.0),
        adc_bit_resolution=16,
    )

    if is_equal:
        assert obj == other_obj
    else:
        assert obj != other_obj


@pytest.mark.parametrize(
    "obj, exp_dict",
    [
        (
            Characteristics(
                quantum_efficiency=0.1,
                charge_to_volt_conversion=0.2,
                pre_amplification=4.4,
                full_well_capacity=10,
            ),
            {
                "quantum_efficiency": 0.1,
                "charge_to_volt_conversion": 0.2,
                "pre_amplification": 4.4,
                "full_well_capacity": 10,
                "adc_bit_resolution": None,
                "adc_voltage_range": None,
            },
        ),
        (
            Characteristics(
                quantum_efficiency=0.1,
                charge_to_volt_conversion=0.2,
                pre_amplification=4.4,
                full_well_capacity=10,
                adc_voltage_range=(0.0, 10.0),
                adc_bit_resolution=16,
            ),
            {
                "quantum_efficiency": 0.1,
                "charge_to_volt_conversion": 0.2,
                "pre_amplification": 4.4,
                "full_well_capacity": 10,
                "adc_voltage_range": (0.0, 10.0),
                "adc_bit_resolution": 16,
            },
        ),
    ],
)
def test_to_and_from_dict(obj, exp_dict):
    """Test methods 'to_dict', 'from_dict'."""
    assert type(obj) == Characteristics

    # Convert from `Characteristics` to a `dict`
    dct = obj.to_dict()
    assert dct == exp_dict

    # Copy 'exp_dict'
    copied_dct = deepcopy(exp_dict)  # create a new dict
    assert copied_dct is not exp_dict
    assert copied_dct == exp_dict

    # Convert from `dict` to `Characteristics`
    other_obj = Characteristics.from_dict(copied_dct)
    assert type(other_obj) == Characteristics
    assert obj == other_obj
    assert obj is not other_obj
    assert copied_dct == exp_dict



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel CosmiX model to generate charge by ionization."""

import bisect
from pathlib import Path
from typing import Callable, Optional

import numpy as np
import pandas as pd
from scipy import interpolate


def sampling_distribution(distribution: np.ndarray) -> float:
    u: float = np.random.random()
    # random_value_from_dist = distribution[bisect.bisect(distribution[:, 1], u) - 1, 0]
    random_value_from_dist = get_xvalue_with_interpolation(
        function_array=distribution, y_value=u
    )

    return random_value_from_dist


def get_xvalue_with_interpolation(function_array: np.ndarray, y_value: float) -> float:
    if y_value <= function_array[0, 1]:
        intpol_x_value: float = function_array[0, 0]
    elif y_value >= function_array[-1, 1]:
        intpol_x_value = function_array[-1, 0]
    else:
        value_bisect: int = bisect.bisect(
            a=function_array[:, 1],  # type: ignore
            x=y_value,
        )
        y_index_bot: int = value_bisect - 1

        y_index_top: int = y_index_bot + 1
        y_value_bot: float = function_array[y_index_bot, 1]
        y_value_top: float = function_array[y_index_top, 1]
        x_value_bot: float = function_array[y_index_bot, 0]
        x_value_top: float = function_array[y_index_top, 0]

        intpol_x_value = x_value_bot + (y_value - y_value_bot) * (
            x_value_top - x_value_bot
        ) / (y_value_top - y_value_bot)

    return intpol_x_value


def get_yvalue_with_interpolation(function_array, x_value):
    x_index_bot = bisect.bisect(function_array[:, 0], x_value) - 1
    x_index_top = x_index_bot + 1
    x_value_bot = function_array[x_index_bot, 0]
    x_value_top = function_array[x_index_top, 0]
    y_value_bot = function_array[x_index_bot, 1]
    y_value_top = function_array[x_index_top, 1]

    intpol_y_value = y_value_bot + (x_value - x_value_bot) * (
        y_value_top - y_value_bot
    ) / (x_value_top - x_value_bot)

    return intpol_y_value


def load_histogram_data(
    file_name: Path,
    hist_type: str,
    skip_rows: Optional[int] = None,
    read_rows: Optional[int] = None,
) -> pd.DataFrame:
    # TODO store count in pandas dataframe as int !!!

    step_size_data = pd.read_csv(
        file_name,
        delimiter="\t",
        names=[hist_type, "counts"],
        usecols=[1, 2],
        skiprows=skip_rows,
        nrows=read_rows,
    )
    return step_size_data


# def load_energy_data(file_name, step_rows):
#     """TBW.
#
#     :param file_name:
#     :param step_rows:
#     :return:
#     """
#     # 'proton_' + energy + '_' + thickness + '_1M.ascii'
#     spectrum_data = pd.read_csv(file_name, delimiter="\t",
#                                 names=["energy", "counts"], usecols=[1, 2], skiprows=step_rows + 8)
#
#     return spectrum_data


def read_data(file_name: Path) -> np.ndarray:
    full_path = file_name.resolve()
    if not full_path.exists():
        raise FileNotFoundError(f"Cannot find file '{full_path}' !")

    data: np.ndarray = np.loadtxt(full_path, "float", "#")
    return data


def interpolate_data(data: np.ndarray) -> Callable[[np.ndarray], np.ndarray]:
    data_function: Callable[[np.ndarray], np.ndarray] = interpolate.interp1d(
        data[:, 0], data[:, 1], kind="linear"
    )
    return data_function



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

import pytest

from pyxel.detectors import CCD, CCDGeometry, Characteristics, Environment
from pyxel.models.charge_transfer import arctic_add


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    return CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )


@pytest.mark.parametrize(
    "trap_densities, trap_release_timescales, exp_exc",
    [
        pytest.param([], [], "Expecting at least one 'trap_density'", id="no trap"),
        pytest.param(
            [1],
            [1, 2],
            "Expecting same number of 'trap_densities' and 'trap_release_timescales'",
            id="no trap",
        ),
    ],
)
def test_arctic_add_bad_inputs(
    ccd_10x10: CCD, trap_densities, trap_release_timescales, exp_exc
):
    """Test function 'arctic_add' with bad inputs."""
    with pytest.raises(ValueError, match=exp_exc):
        arctic_add(
            detector=ccd_10x10,
            well_fill_power=1.0,
            trap_densities=trap_densities,
            trap_release_timescales=trap_release_timescales,
            express=0,
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Detector class."""
import collections
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

import numpy as np

from pyxel import __version__, backends
from pyxel.data_structure import (
    Charge,
    Image,
    Persistence,
    Photon,
    Pixel,
    Scene,
    Signal,
    SimplePersistence,
)
from pyxel.detectors import Environment, ReadoutProperties
from pyxel.util.memory import get_size, memory_usage_details

if TYPE_CHECKING:
    import xarray as xr
    from datatree import DataTree


__all__ = ["Detector"]


# TODO: Add methods to save/load a `Detector` instance to the filesystem. See #329
class Detector:
    """The detector class."""

    def __init__(self, environment: Optional[Environment] = None):
        self.environment: Environment = environment or Environment()

        self.header: dict[str, object] = collections.OrderedDict()

        self._photon: Optional[Photon] = None
        self._scene: Optional[Scene] = None
        self._charge: Optional[Charge] = None
        self._pixel: Optional[Pixel] = None
        self._signal: Optional[Signal] = None
        self._image: Optional[Image] = None
        self._data: Optional[DataTree] = None

        # This will be the memory of the detector where trapped charges will be saved
        self._memory: dict = {}
        self._persistence: Optional[Union[Persistence, SimplePersistence]] = None

        self.input_image: Optional[np.ndarray] = None
        self._output_dir: Optional[Path] = None  # TODO: See #330

        self._readout_properties: Optional["ReadoutProperties"] = None

        self._numbytes = get_size(self)

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, Detector)
            and self._photon == other._photon
            and self._scene == other._scene
            and self._charge == other._charge
            and self._pixel == other._pixel
            and self._signal == other._signal
            and self._image == other._image
            and (
                (self._data is other._data is None)
                or (
                    self._data is not None
                    and other._data is not None
                    and self._data.equals(other._data)
                )
            )
        )

    @property
    def geometry(self):
        """TBW."""
        raise NotImplementedError

    @property
    def characteristics(self):
        """TBW."""
        raise NotImplementedError

    @property
    def photon(self) -> Photon:
        """TBW."""
        if not self._photon:
            raise RuntimeError("Photon array is not initialized ! ")
        return self._photon

    @photon.setter
    def photon(self, obj: Photon) -> None:
        self._photon = obj

    @property
    def scene(self) -> Scene:
        """TBW."""
        if not self._scene:
            raise RuntimeError("Scene object is not initialized ! ")
        return self._scene

    @scene.setter
    def scene(self, obj: Scene) -> None:
        """TBW."""
        self._scene = obj

    @property
    def charge(self) -> Charge:
        """TBW."""
        if not self._charge:
            raise RuntimeError("'charge' not initialized.")

        return self._charge

    @property
    def pixel(self) -> Pixel:
        """TBW."""
        if not self._pixel:
            raise RuntimeError("'pixel' not initialized.")

        return self._pixel

    @property
    def signal(self) -> Signal:
        """TBW."""
        if not self._signal:
            raise RuntimeError("'signal' not initialized.")

        return self._signal

    @property
    def image(self) -> Image:
        """TBW."""
        if not self._image:
            raise RuntimeError("'image' not initialized.")

        return self._image

    @property
    def data(self) -> "DataTree":
        """TBW."""
        if self._data is None:
            raise RuntimeError("'data' not initialized.")

        return self._data

    def to_xarray(self) -> "xr.Dataset":
        """Create a new ``Dataset`` from all data containers.

        Examples
        --------
        >>> detector.to_xarray()
        <xarray.Dataset>
        Dimensions:  (y: 100, x: 100)
        Coordinates:
          * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
          * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        Data variables:
            photon   (y, x) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0 0.0
            pixel    (y, x) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0 0.0
            signal   (y, x) float64 0.0 0.0 0.0 0.0 0.0 0.0 ... 0.0 0.0 0.0 0.0 0.0 0.0
            image    (y, x) uint64 0 0 0 0 0 0 0 0 0 0 0 0 0 ... 0 0 0 0 0 0 0 0 0 0 0 0
        Attributes:
            detector:       CCD
            pyxel version:  1.5
        """
        import xarray as xr

        ds = xr.Dataset()
        ds["photon"] = self.photon.to_xarray()
        # ds["scene"] = self.scene.to_xarray()
        ds["pixel"] = self.pixel.to_xarray()
        ds["signal"] = self.signal.to_xarray()
        ds["image"] = self.image.to_xarray()
        ds["charge"] = self.charge.to_xarray()

        ds.attrs.update({"detector": type(self).__name__, "pyxel version": __version__})

        return ds

    def reset(self) -> None:
        """TBW."""
        from datatree import DataTree

        self._photon = Photon(geo=self.geometry)
        self._scene = Scene()
        self._charge = Charge(geo=self.geometry)
        self._pixel = Pixel(geo=self.geometry)
        self._signal = Signal(geo=self.geometry)
        self._image = Image(geo=self.geometry)

        self._data = DataTree()

    def empty(self, empty_all: bool = True) -> None:
        """Empty the data in the detector."""
        if self._photon:
            self.photon.array *= 0

        self._scene = Scene()

        if self._signal:
            self.signal.array *= 0
        if self._image:
            self.image.array *= 0
        if self._charge:
            self._charge.empty()
        if empty_all:
            if self._pixel:
                self.pixel.array *= 0

    # TODO: Set an `Output` object ? Is it really needed ? See #330
    def set_output_dir(self, path: Union[str, Path]) -> None:
        """Set output directory path."""
        self._output_dir = Path(path)

    # TODO: Set an `Output` object ? Is it really needed ? See #330
    @property
    def output_dir(self) -> Path:
        """Output directory path."""
        if self._output_dir is None:
            raise RuntimeError("'output_dir' is not initialized.")

        return self._output_dir

    def set_readout(
        self,
        times: Union[Sequence[float], np.ndarray],
        start_time: float = 0.0,
        non_destructive: bool = False,
    ) -> None:
        """Set readout sampling properties.

        Parameters
        ----------
        times : Sequence[Number]
            A sequence of numeric values representing the sampling times for the readout simulation.
        start_time : float, optional. Default: 0.0
            A float representing the starting time of the readout simulation.
            The readout time(s) should be greater that this ``start_time``.
        non_destructive : bool, optional. Default: False
            A boolean flag indicating whether the readout simulation is non-destructive.
            If set to ``True``, the readout process will not modify the underlying data.

        Examples
        --------
        >>> detector.set_readout(times=[1, 2, 4, 7, 10], start_time=0.0)
        """
        self._readout_properties = ReadoutProperties(
            times=times,
            start_time=start_time,
            non_destructive=non_destructive,
        )

    @property
    def readout_properties(self) -> ReadoutProperties:
        """Return current ``ReadoutProperties``."""
        if self._readout_properties is None:
            raise ValueError("No readout defined.")

        return self._readout_properties

    @property
    def time(self) -> float:
        """TBW."""
        return self.readout_properties.time

    @time.setter
    def time(self, value: float) -> None:
        """TBW."""
        self.readout_properties.time = value

    @property
    def start_time(self) -> float:
        """TBW."""
        return self.readout_properties.start_time

    @start_time.setter
    def start_time(self, value: float) -> None:
        """TBW."""
        self.readout_properties.start_time = value

    @property
    def absolute_time(self) -> float:
        """TBW."""
        return self.readout_properties.absolute_time

    @property
    def time_step(self) -> float:
        """TBW."""
        return self.readout_properties.time_step

    @time_step.setter
    def time_step(self, value: float) -> None:
        """TBW."""
        self.readout_properties.time_step = value

    @property
    def times_linear(self) -> bool:
        """TBW."""
        return self.readout_properties.times_linear

    @property
    def num_steps(self) -> int:
        """TBW."""
        return self.readout_properties.num_steps

    @property
    def pipeline_count(self) -> int:
        """TBW."""
        return self.readout_properties.pipeline_count

    @pipeline_count.setter
    def pipeline_count(self, value: int) -> None:
        """TBW."""
        self.readout_properties.pipeline_count = value

    @property
    def is_first_readout(self) -> bool:
        """Check if this is the first readout time."""
        return self.readout_properties.is_first_readout

    @property
    def is_last_readout(self) -> bool:
        """Check if this is the last readout time."""
        return bool(self.pipeline_count == (self.num_steps - 1))

    @property
    def read_out(self) -> bool:
        """TBW."""
        return self.readout_properties.read_out

    @read_out.setter
    def read_out(self, value: bool) -> None:
        """TBW."""
        self.readout_properties.read_out = value

    @property
    def is_dynamic(self) -> bool:
        """Return if detector is dynamic (time dependent) or not.

        By default it is not dynamic.
        """
        if self._readout_properties is not None:
            return True

        return False

    @property
    def non_destructive_readout(self) -> bool:
        """Return if detector readout mode is destructive or integrating.

        By default it is destructive (non-integrating).
        """
        return self.readout_properties.non_destructive

    def has_persistence(self) -> bool:
        """TBW."""
        if self._persistence is not None:
            return True

        return False

    @property
    def persistence(self) -> Union[Persistence, SimplePersistence]:
        """TBW."""
        if self._persistence is not None:
            return self._persistence
        else:
            raise RuntimeError("'persistence' not initialized.")

    @persistence.setter
    def persistence(self, value: Union[Persistence, SimplePersistence]) -> None:
        """TBW."""
        if not isinstance(value, (Persistence, SimplePersistence)):
            raise TypeError(
                "Expecting Persistence or SimplePersistence type to set detector persistence."
            )
        self._persistence = value

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes

    def memory_usage(
        self, print_result: bool = True, human_readable: bool = True
    ) -> dict:
        """TBW.

        Returns
        -------
        dict
            Dictionary of attribute memory usage
        """
        attributes = [
            "_photon",
            "_scene",
            "_charge",
            "_pixel",
            "_signal",
            "_image",
            "material",
            "environment",
            "_geometry",
            "_characteristics",
        ]

        return memory_usage_details(
            self, attributes, print_result=print_result, human_readable=human_readable
        )

    @classmethod
    def load(cls, filename: Union[str, Path]) -> "Detector":
        """Load a detector object from a filename.

        Parameters
        ----------
        filename : str or Path

        Returns
        -------
        Detector
            A new ``Detector`` object.

        Raises
        ------
        FileNotFoundError
            If ``filename`` is not found.
        ValueError
            If the extension of filename is not recognized.
        """
        full_filename = Path(filename).resolve()
        if not full_filename.exists():
            raise FileNotFoundError(f"Filename '{filename}' does not exist !")

        extension: str = full_filename.suffix

        if extension in (".h5", ".hdf5", ".hdf"):
            return cls.from_hdf5(filename)
        elif extension == ".asdf":
            return cls.from_asdf(filename)
        else:
            raise ValueError(f"Unknown extension {extension!r}.")

    def save(self, filename: Union[str, Path]) -> None:
        """Save a detector object into a filename.

        Parameters
        ----------
        filename : str or Path

        Raises
        ------
        ValueError
            If the extension of filename is not recognized.
        """
        full_filename = Path(filename).resolve()
        extension: str = full_filename.suffix

        if extension in (".h5", ".hdf5", ".hdf"):
            return self.to_hdf5(filename)
        elif extension == ".asdf":
            return self.to_asdf(filename)
        else:
            raise ValueError(f"Unknown extension {extension!r}.")

    # TODO: Move this to another place. See #241
    def to_hdf5(self, filename: Union[str, Path]) -> None:
        """Write the detector content to a :term:`HDF5` file.

        The HDF5 file has the following structure:

        .. code-block:: bash

            filename.h5  (4 objects, 3 attributes)
            │   ├── pyxel-version  1.0.0+161.g659eec86
            │   ├── type  CCD
            │   └── version  1
            ├── geometry  (5 objects)
            │   ├── col  (), int64
            │   ├── pixel_horz_size  (), float64
            │   ├── pixel_vert_size  (), float64
            │   ├── row  (), int64
            │   └── total_thickness  (), float64
            ├── environment  (1 object)
            │   └── temperature  (), float64
            ├── characteristics  (4 objects)
            │   ├── charge_to_volt_conversion  (), float64
            │   ├── full_well_capacity  (), int64
            │   ├── pre_amplification  (), float64
            │   └── quantum_efficiency  (), float64
            └── data  (5 objects)
                ├── charge  (2 objects, 2 attributes)
                │   ├── name  Charge
                │   ├── unit  electron
                │   ├── array  (100, 120), float64
                │   └── frame  (13 objects, 1 attribute)
                │       ├── type  DataFrame
                │       ├── charge  (0,), float64
                │       ├── energy  (0,), float64
                │       ├── init_energy  (0,), float64
                │       ├── init_pos_hor  (0,), float64
                │       ├── init_pos_ver  (0,), float64
                │       ├── init_pos_z  (0,), float64
                │       ├── number  (0,), float64
                │       ├── position_hor  (0,), float64
                │       ├── position_ver  (0,), float64
                │       ├── position_z  (0,), float64
                │       ├── velocity_hor  (0,), float64
                │       ├── velocity_ver  (0,), float64
                │       └── velocity_z  (0,), float64
                ├── image  (100, 120), uint64
                │   └── name  Image
                ├── photon  (100, 120), float64
                ├── pixel  (100, 120), float64
                └── signal  (100, 120), float64

        Parameters
        ----------
        filename : str or Path

        Notes
        -----
        You can find more information in the 'how-to' guide section.

        Examples
        --------
        >>> from pyxel.detectors import CCD
        >>> detector = CCD(...)

        >>> detector.to_hdf5("ccd.h5")
        """
        dct: Mapping = self.to_dict()
        backends.to_hdf5(filename=filename, dct=dct)

    @classmethod
    def from_hdf5(cls, filename: Union[str, Path]) -> "Detector":
        """Load a detector object from a :term:`HDF5` file.

        Parameters
        ----------
        filename : str or Path

        Examples
        --------
        >>> detector = Detector.from_hdf5("ccd.h5")
        >>> detector
        CCD(...)
        """
        dct: Mapping[str, Any]
        with backends.from_hdf5(filename) as dct:
            obj: Detector = cls.from_dict(dct)
            return obj

    def to_asdf(self, filename: Union[str, Path]) -> None:
        """Write the detector content to a :term:`ASDF` file.

        The ASDF file has the following structure:

        .. code-block:: bash

             root (AsdfObject)
             ├─version (int): 1
             ├─type (str): CCD
             ├─properties (dict)
             │ ├─geometry (dict)
             │ │ ├─row (int): 4
             │ │ ├─col (int): 5
             │ │ ├─total_thickness (NoneType): None
             │ │ ├─pixel_vert_size (NoneType): None
             │ │ └─pixel_horz_size (NoneType): None
             │ ├─environment (dict)
             │ │ └─temperature (NoneType): None
             │ └─characteristics (dict)
             │   ├─quantum_efficiency (NoneType): None
             │   ├─charge_to_volt_conversion (NoneType): None
             │   └─4 not shown
             └─data (dict)
               ├─photon (ndarray): shape=(4, 5), dtype=float64
               ├─scene (NoneType): None
               ├─pixel (ndarray): shape=(4, 5), dtype=float64
               ├─signal (ndarray): shape=(4, 5), dtype=float64
               ├─image (ndarray): shape=(4, 5), dtype=uint64
               └─charge (dict) ...

        Parameters
        ----------
        filename : str or Path

        Notes
        -----
        You can find more information in the 'how-to' guide section.

        Examples
        --------
        >>> from pyxel.detectors import CCD
        >>> detector = CCD(...)

        >>> detector.to_asdf("ccd.asdf")

        >>> import asdf
        >>> af = asdf.open("ccd_asdf")
        >>> af["type"]
        'CCD'
        >>> af.info()
        """
        dct: Mapping = self.to_dict()
        backends.to_asdf(filename=filename, dct=dct)

    @classmethod
    def from_asdf(cls, filename: Union[str, Path]) -> "Detector":
        """Load a detector object from a :term:`ASDF` file.

        Parameters
        ----------
        filename : str or Path

        Examples
        --------
        >>> detector = Detector.from_asdf("ccd.asdf")
        >>> detector
        CCD(...)
        """
        with backends.from_asdf(filename) as dct:
            detector: Detector = cls.from_dict(dct)

        return detector

    def to_dict(self) -> Mapping:
        """Convert a `Detector` to a `dict`."""
        raise NotImplementedError

    # TODO: Replace `-> 'Detector'` by `Union[CCD, CMOS, MKID]`
    @classmethod
    def from_dict(cls, dct: Mapping) -> "Detector":
        """Create a new instance of a `Detector` from a `dict`."""
        # TODO: This is a simplistic implementation. Improve this.
        if dct["type"] == "CCD":
            from pyxel.detectors import CCD  # Imported here to avoid circular import

            return CCD.from_dict(dct)

        elif dct["type"] == "CMOS":
            from pyxel.detectors import CMOS

            return CMOS.from_dict(dct)

        elif dct["type"] == "MKID":
            from pyxel.detectors import MKID

            return MKID.from_dict(dct)

        elif dct["type"] == "APD":
            from pyxel.detectors import APD

            return APD.from_dict(dct)

        else:
            raise NotImplementedError(f"Unknown type: {dct['type']!r}")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""CLI to run Pyxel."""
import logging
import sys
import time
from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Union

import click
import dask
import pandas as pd

from pyxel import Configuration
from pyxel import __version__ as version
from pyxel import load, outputs, save
from pyxel.detectors import APD, CCD, CMOS, MKID, Detector
from pyxel.exposure import Exposure
from pyxel.observation import Observation, ObservationResult
from pyxel.pipelines import DetectionPipeline, Processor
from pyxel.util import create_model, download_examples

if TYPE_CHECKING:
    import xarray as xr
    from datatree import DataTree

    from pyxel.calibration import Calibration
    from pyxel.outputs import CalibrationOutputs, ExposureOutputs, ObservationOutputs


# TODO: This function will be deprecated (see #563)
def exposure_mode(
    exposure: "Exposure",
    detector: Detector,
    pipeline: "DetectionPipeline",
) -> "xr.Dataset":
    """Run an 'exposure' pipeline.

    For more information, see :ref:`exposure_mode`.

    Parameters
    ----------
    exposure: Exposure
    detector: Detector
    pipeline: DetectionPipeline

    Returns
    -------
    Dataset
        An multi-dimensional array database from `xarray <https://xarray.pydata.org>`_.

    Examples
    --------
    Load a configuration file

    >>> import pyxel
    >>> config = pyxel.load("configuration.yaml")
    >>> config
    Configuration(...)

    Run an exposure pipeline

    >>> dataset = pyxel.exposure_mode(
    ...     exposure=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )
    >>> dataset
    <xarray.Dataset>
    Dimensions:       (readout_time: 1, y: 450, x: 450)
    Coordinates:
      * readout_time  (readout_time) int64 1
      * y             (y) int64 0 1 2 3 4 5 6 7 ... 442 443 444 445 446 447 448 449
      * x             (x) int64 0 1 2 3 4 5 6 7 ... 442 443 444 445 446 447 448 449
    Data variables:
        image         (readout_time, y, x) uint16 9475 9089 8912 ... 9226 9584 10079
        signal        (readout_time, y, x) float64 3.159 3.03 2.971 ... 3.195 3.36
        pixel         (readout_time, y, x) float64 1.053e+03 1.01e+03 ... 1.12e+03
    """

    logging.info("Mode: Exposure")

    exposure_outputs: ExposureOutputs = exposure.outputs
    detector.set_output_dir(exposure_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    result: xr.Dataset = exposure.run_exposure(processor=processor)

    if exposure_outputs.save_exposure_data:
        exposure_outputs.save_exposure_outputs(dataset=result)

    return result


def _run_exposure_mode(
    exposure: "Exposure",
    detector: Detector,
    pipeline: "DetectionPipeline",
    with_intermediate_steps: bool,
) -> "DataTree":
    """Run an 'exposure' pipeline.

    For more information, see :ref:`exposure_mode`.

    Parameters
    ----------
    exposure : Exposure
    detector : Detector
    pipeline : DetectionPipeline
    with_intermediate_steps : bool

    Returns
    -------
    DataTree
        An multi-dimensional tree of arrays.

    Examples
    --------
    Load a configuration file

    >>> import pyxel
    >>> config = pyxel.load("configuration.yaml")
    >>> config
    Configuration(...)

    Run an exposure pipeline

    >>> data_tree = pyxel._run_exposure_mode(
    ...     exposure=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )
    >>> data_tree
    DataTree('None', parent=None)
    │   Dimensions:  (time: 54, y: 100, x: 100)
    │   Coordinates:
    │     * time     (time) float64 0.02 0.06 0.12 0.2 0.3 ... 113.0 117.8 122.7 127.7
    │     * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
    │     * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
    │   Data variables:
    │       photon   (time, y, x) float64 102.0 108.0 79.0 ... 2.513e+04 2.523e+04
    │       charge   (time, y, x) float64 201.0 193.0 173.0 ... 2.523e+04 2.532e+04
    │       pixel    (time, y, x) float64 94.68 98.18 71.82 ... 2.388e+04 2.418e+04
    │       signal   (time, y, x) float64 0.001176 0.00129 0.0007866 ... 0.2946 0.2982
    │       image    (time, y, x) float64 20.0 22.0 13.0 ... 4.826e+03 4.887e+03
    │   Attributes:
    │       pyxel version:  1.9.1+104.g9da11bb2.dirty
    │       running mode:   Exposure
    └── DataTree('data')
        └── DataTree('mean_variance')
            └── DataTree('image')
                    Dimensions:   (mean: 54)
                    Coordinates:
                      * mean      (mean) float64 19.58 38.7 57.83 ... 4.586e+03 4.681e+03 4.776e+03
                    Data variables:
                        variance  (mean) float64 5.958 10.28 14.82 ... 1.25e+04 1.3e+04 1.348e+04
    """

    logging.info("Mode: Exposure")

    exposure_outputs: ExposureOutputs = exposure.outputs
    detector.set_output_dir(exposure_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    result: DataTree = exposure.run_exposure_new(
        processor=processor,
        with_intermediate_steps=with_intermediate_steps,
    )

    if exposure_outputs.save_exposure_data:
        exposure_outputs.save_exposure_outputs(dataset=result)

    return result


# TODO: This function will be deprecated (see #563)
def observation_mode(
    observation: "Observation",
    detector: Detector,
    pipeline: "DetectionPipeline",
) -> "ObservationResult":
    """Run an 'observation' pipeline.

    For more information, see :ref:`observation_mode`.

    Parameters
    ----------
    observation: Observation
    detector: Detector
    pipeline: DetectionPipeline

    Returns
    -------
    ObservationResult
        Result.

    Examples
    --------
    Load a configuration file

    >>> import pyxel
    >>> config = pyxel.load("configuration.yaml")
    >>> config
    Configuration(...)

    Run an observation pipeline

    >>> result = pyxel.observation_mode(
    ...     exposure=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )
    >>> result
    ObservationResult(...)
    """
    logging.info("Mode: Observation")

    observation_outputs: ObservationOutputs = observation.outputs
    detector.set_output_dir(observation_outputs.output_dir)  # TODO: Remove this

    # TODO: This should be done during initializing of object `Configuration`
    # parametric_outputs.params_func(parametric)

    processor = Processor(detector=detector, pipeline=pipeline)

    result: ObservationResult = observation.run_observation(processor=processor)

    if observation_outputs.save_observation_data:
        observation_outputs.save_observation_datasets(
            result=result, mode=observation.parameter_mode
        )

    return result


# TODO: This function will be deprecated (see #563)
def calibration_mode(
    calibration: "Calibration",
    detector: Detector,
    pipeline: "DetectionPipeline",
    compute_and_save: bool = True,
) -> tuple["xr.Dataset", pd.DataFrame, pd.DataFrame, Sequence]:
    """Run a 'calibration' pipeline.

    For more information, see :ref:`calibration_mode`.

    Parameters
    ----------
    calibration: Calibration
    detector: Detector
    pipeline: DetectionPipeline
    compute_and_save: bool

    Returns
    -------
    tuple of Dataset, DataFrame, DataFrame, Sequence

    Examples
    --------
    Load a configuration file

    >>> import pyxel
    >>> config = pyxel.load("configuration.yaml")
    >>> config
    Configuration(...)

    Run a calibration pipeline

    >>> ds, processors, logs, filenames = pyxel.calibration_mode(
    ...     exposure=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )

    >>> ds
    <xarray.Dataset>
    Dimensions:              (island: 2, evolution: 2, param_id: 4, id_processor: 1, readout_time: 1, y: 100, x: 100)
    Coordinates:
      * island               (island) int64 1 2
      * evolution            (evolution) int64 1 2
      * id_processor         (id_processor) int64 0
      * readout_time         (readout_time) int64 1
      * y                    (y) int64 0 1 2 3 4 5 6 7 8 ... 92 93 94 95 96 97 98 99
      * x                    (x) int64 0 1 2 3 4 5 6 7 8 ... 92 93 94 95 96 97 98 99
      * param_id             (param_id) int64 0 1 2 3
    Data variables:
        champion_fitness     (island, evolution) float64 4.759e+06 ... 4.533e+06
        champion_decision    (island, evolution, param_id) float64 0.08016 ... 7....
        champion_parameters  (island, evolution, param_id) float64 0.08016 ... 7....
        simulated_image      (island, id_processor, readout_time, y, x) uint16 48...
        simulated_signal     (island, id_processor, readout_time, y, x) float64 4...
        simulated_pixel      (island, id_processor, readout_time, y, x) float64 6...
        target               (id_processor, y, x) >f8 4.834e+03 ... 4.865e+03
    Attributes:
        num_islands:      2
        population_size:  20
        num_evolutions:   2
        generations:      5
        topology:         unconnected
        result_type:      image

    >>> processors
           island  id_processor                                          processor
    0       0             0  Delayed('apply_parameters-c5da1649-766f-4ecb-a...
    1       1             0  Delayed('apply_parameters-c16f998f-f52f-4beb-b...

    >>> logs
        num_generations  ...  global_num_generations
    0                 1  ...                       1
    1                 2  ...                       2
    2                 3  ...                       3
    3                 4  ...                       4
    4                 5  ...                       5
    ..              ...  ...                     ...
    15                1  ...                       6
    16                2  ...                       7
    17                3  ...                       8
    18                4  ...                       9
    19                5  ...                      10

    >>> filenames
    []
    """
    # Late import to speedup start-up time
    from pyxel.calibration import CalibrationResult

    logging.info("Mode: Calibration")

    calibration_outputs: CalibrationOutputs = calibration.outputs
    detector.set_output_dir(calibration_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    ds_results, df_processors, df_all_logs = calibration.run_calibration(
        processor=processor, output_dir=calibration_outputs.output_dir
    )

    # TODO: Save the processors from 'df_processors'
    # TODO: Generate plots from 'ds_results'

    # TODO: Do something with 'df_all_logs' ?

    # TODO: create 'output' object with .calibration_outputs
    # TODO: use 'fitting.get_simulated_data' ==> np.ndarray

    # geometry = processor.detector.geometry
    # calibration.post_processing(
    #     champions=champions,
    #     output=calibration_outputs,
    #     row=geometry.row,
    #     col=geometry.col,
    # )
    filenames = calibration._post_processing(
        ds=ds_results, df_processors=df_processors, output=calibration_outputs
    )

    if compute_and_save:
        computed_ds, df_processors, df_logs, filenames = dask.compute(
            ds_results, df_processors, df_all_logs, filenames
        )

        if calibration_outputs.save_calibration_data:
            calibration_outputs.save_calibration_outputs(
                dataset=computed_ds, logs=df_logs
            )
            print(f"Saved calibration outputs to {calibration_outputs.output_dir}")

        result = CalibrationResult(
            dataset=computed_ds,
            processors=df_processors,
            logs=df_logs,
            filenames=filenames,
        )

    else:
        result = CalibrationResult(
            dataset=ds_results,
            processors=df_processors,
            logs=df_all_logs,
            filenames=filenames,
        )

    return result


def _run_calibration_mode(
    calibration: "Calibration",
    detector: Detector,
    pipeline: "DetectionPipeline",
) -> "DataTree":
    """Run a 'Calibration' pipeline.

    Parameters
    ----------
    calibration : Calibration
    detector : Detector
    pipeline : DetectionPipeline

    Notes
    -----
    This is a 'private' function called by 'run_mode'.

    Returns
    -------
    DataTree

    Examples
    --------
    >>> data_tree = _run_calibration_mode(calibration=..., detector=..., pipeline=...)
    >>> data_tree
    DataTree('None', parent=None)
    │   Dimensions:              (evolution: 5, island: 1, param_id: 4, individual: 10,
    │                             processor: 10, readout_time: 1, y: 235, x: 1)
    │   Coordinates:
    │     * evolution            (evolution) int64 0 1 2 3 4
    │     * island               (island) int64 0
    │     * param_id             (param_id) int64 0 1 2 3
    │     * individual           (individual) int64 0 1 2 3 4 5 6 7 8 9
    │     * processor            (processor) int64 0 1 2 3 4 5 6 7 8 9
    │     * readout_time         (readout_time) int64 1
    │     * y                    (y) int64 2065 2066 2067 2068 ... 2296 2297 2298 2299
    │     * x                    (x) int64 0
    │   Data variables:
    │       champion_fitness     (island, evolution) float64 3.271e+06 ... 4.641e+05
    │       champion_decision    (island, evolution, param_id) float64 -2.224 ... 3.662
    │       champion_parameters  (island, evolution, param_id) float64 0.00597 ... 4....
    │       best_fitness         (island, evolution, individual) float64 3.271e+06 .....
    │       best_decision        (island, evolution, individual, param_id) float64 -2...
    │       best_parameters      (island, evolution, individual, param_id) float64 0....
    │       simulated_photon     (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       simulated_charge     (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       simulated_pixel      (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       simulated_signal     (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       simulated_image      (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       target               (processor, y, x) float64 13.75 0.4567 ... 0.2293 0.375
    │   Attributes:
    │       num_islands:      1
    │       population_size:  10
    │       num_evolutions:   5
    │       generations:      1
    │       topology:         fully_connected
    │       result_type:      pixel
    └── DataTree('full_size')
            Dimensions:           (island: 1, processor: 10, readout_time: 1, y: 2300, x: 1)
            Coordinates:
              * island            (island) int64 0
              * processor         (processor) int64 0 1 2 3 4 5 6 7 8 9
              * readout_time      (readout_time) int64 1
              * y                 (y) int64 0 1 2 3 4 5 6 ... 2294 2295 2296 2297 2298 2299
              * x                 (x) int64 0
            Data variables:
                simulated_photon  (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                simulated_charge  (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                simulated_pixel   (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                simulated_signal  (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                simulated_image   (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                target            (processor, y, x) float64 0.0 0.4285 ... 0.2293 0.375
    """
    logging.info("Mode: Calibration")

    calibration_outputs: CalibrationOutputs = calibration.outputs
    detector.set_output_dir(calibration_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    data_tree = calibration.run_calibration_new(
        processor=processor,
        output_dir=calibration_outputs.output_dir,
    )

    return data_tree


def _run_observation_mode(
    observation: Observation,
    detector: Detector,
    pipeline: DetectionPipeline,
) -> "DataTree":
    logging.info("Mode: Observation")

    observation_outputs: ObservationOutputs = observation.outputs
    detector.set_output_dir(observation_outputs.output_dir)  # TODO: Remove this

    processor = Processor(detector=detector, pipeline=pipeline)

    result = observation.run_observation_datatree(processor=processor)

    if observation_outputs.save_observation_data:
        raise NotImplementedError
    #     observation_outputs.save_observation_datasets(
    #         result=result, mode=observation.parameter_mode
    #     )

    return result


def run_mode(
    mode: Union[Exposure, Observation, "Calibration"],
    detector: Detector,
    pipeline: DetectionPipeline,
    with_intermediate_steps: bool = False,
) -> "DataTree":
    """Run a pipeline.

    Parameters
    ----------
    mode : Exposure, Observation or Calibration
        Mode to execute.
    detector : Detector
        This object is the container for all the data used for the models.
    pipeline : DetectionPipeline
        This is the core algorithm of Pyxel. This pipeline contains all the models to run.
    with_intermediate_steps : bool, default: False
        Add all intermediate steps into the results as a ``DataTree``. This mode is used for debugging.


    Notes
    -----
    Parameter ``with_intermediate_steps`` is not (yet) stable and may change in the future.

    Returns
    -------
    DataTree

    Raises
    ------
    TypeError
        Raised if the ``mode`` is not valid.

    NotImplementedError
        Raised if parameter ``with_intermediate_steps`` is activated and `mode` is not an ``Exposure`` object.

    Examples
    --------
    Load a configuration file

    >>> import pyxel
    >>> config = pyxel.load("exposure_configuration.yaml")
    >>> config

     Run a 'Exposure' pipeline

    >>> data = pyxel.run_mode(
    ...     mode=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )
    >>> data
    DataTree('None', parent=None)
    │   Dimensions:  (time: 54, y: 100, x: 100)
    │   Coordinates:
    │     * time     (time) float64 0.02 0.06 0.12 0.2 0.3 ... 113.0 117.8 122.7 127.7
    │     * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
    │     * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
    │   Data variables:
    │       photon   (time, y, x) float64 85.0 120.0 109.0 ... 2.533e+04 2.51e+04
    │       charge   (time, y, x) float64 201.0 196.0 202.0 ... 2.543e+04 2.52e+04
    │       pixel    (time, y, x) float64 77.38 110.0 99.09 ... 2.406e+04 2.406e+04
    │       signal   (time, y, x) float64 0.0009377 0.001322 0.00133 ... 0.2968 0.2968
    │       image    (time, y, x) float64 16.0 22.0 22.0 ... 4.863e+03 4.863e+03
    │   Attributes:
    │       pyxel version:  1.9.1+104.g9da11bb2
    │       running mode:   Exposure
    ├── DataTree('scene')
    │   └── DataTree('list')
    │       └── DataTree('0')
    │               Dimensions:     (ref: 345, wavelength: 343)
    │               Coordinates:
    │                 * ref         (ref) int64 0 1 2 3 4 5 6 7 ... 337 338 339 340 341 342 343 344
    │                 * wavelength  (wavelength) float64 336.0 338.0 340.0 ... 1.018e+03 1.02e+03
    │               Data variables:
    │                   x           (ref) float64 2.057e+05 2.058e+05 ... 2.031e+05 2.03e+05
    │                   y           (ref) float64 8.575e+04 8.58e+04 ... 8.795e+04 8.807e+04
    │                   weight      (ref) float64 11.49 14.13 15.22 14.56 ... 15.21 11.51 8.727
    │                   flux        (ref, wavelength) float64 0.03769 0.04137 ... 1.813 1.896
    └── DataTree('data')
        ├── DataTree('mean_variance')
        │   └── DataTree('image')
        │           Dimensions:   (mean: 54)
        │           Coordinates:
        │             * mean      (mean) float64 19.64 38.7 57.77 ... 4.586e+03 4.682e+03 4.777e+03
        │           Data variables:
        │               variance  (mean) float64 5.893 10.36 15.13 ... 1.235e+04 1.297e+04 1.342e+04
        └── DataTree('statistics')
            └── DataTree('pixel')
                    Dimensions:  (time: 54)
                    Coordinates:
                      * time     (time) float64 0.02 0.06 0.12 0.2 0.3 ... 113.0 117.8 122.7 127.7
                    Data variables:
                        var      (time) float64 92.4 197.8 317.2 ... 3.027e+05 3.175e+05 3.286e+05
                        mean     (time) float64 94.64 189.1 283.5 ... 2.269e+04 2.316e+04 2.363e+04
                        min      (time) float64 63.39 134.9 220.3 ... 2.135e+04 2.193e+04 2.24e+04
                        max      (time) float64 134.8 248.1 359.7 ... 2.522e+04 2.569e+04 2.64e+04
                        count    (time) float64 1e+04 1e+04 1e+04 1e+04 ... 1e+04 1e+04 1e+04 1e+04

    Run a 'Calibration' pipeline

    >>> config = pyxel.load("calibration_configuration.yaml")
    >>> data = pyxel.run_mode(
    ...     mode=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ... )
    >>> data
    DataTree('None', parent=None)
    │   Dimensions:              (evolution: 5, island: 1, param_id: 4, individual: 10,
    │                             processor: 10, readout_time: 1, y: 235, x: 1)
    │   Coordinates:
    │     * evolution            (evolution) int64 0 1 2 3 4
    │     * island               (island) int64 0
    │     * param_id             (param_id) int64 0 1 2 3
    │     * individual           (individual) int64 0 1 2 3 4 5 6 7 8 9
    │     * processor            (processor) int64 0 1 2 3 4 5 6 7 8 9
    │     * readout_time         (readout_time) int64 1
    │     * y                    (y) int64 2065 2066 2067 2068 ... 2296 2297 2298 2299
    │     * x                    (x) int64 0
    │   Data variables:
    │       champion_fitness     (island, evolution) float64 3.271e+06 ... 4.641e+05
    │       champion_decision    (island, evolution, param_id) float64 -2.224 ... 3.662
    │       champion_parameters  (island, evolution, param_id) float64 0.00597 ... 4....
    │       best_fitness         (island, evolution, individual) float64 3.271e+06 .....
    │       best_decision        (island, evolution, individual, param_id) float64 -2...
    │       best_parameters      (island, evolution, individual, param_id) float64 0....
    │       simulated_photon     (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       simulated_charge     (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       simulated_pixel      (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       simulated_signal     (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       simulated_image      (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 235, 1), meta=np.ndarray>
    │       target               (processor, y, x) float64 13.75 0.4567 ... 0.2293 0.375
    │   Attributes:
    │       num_islands:      1
    │       population_size:  10
    │       num_evolutions:   5
    │       generations:      1
    │       topology:         fully_connected
    │       result_type:      pixel
    └── DataTree('full_size')
            Dimensions:           (island: 1, processor: 10, readout_time: 1, y: 2300, x: 1)
            Coordinates:
              * island            (island) int64 0
              * processor         (processor) int64 0 1 2 3 4 5 6 7 8 9
              * readout_time      (readout_time) int64 1
              * y                 (y) int64 0 1 2 3 4 5 6 ... 2294 2295 2296 2297 2298 2299
              * x                 (x) int64 0
            Data variables:
                simulated_photon  (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                simulated_charge  (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                simulated_pixel   (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                simulated_signal  (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                simulated_image   (island, processor, readout_time, y, x) float64 dask.array<chunksize=(1, 1, 1, 2300, 1), meta=np.ndarray>
                target            (processor, y, x) float64 0.0 0.4285 ... 0.2293 0.375

    Run a pipeline with all intermediate steps

    >>> results = pyxel.run_mode(
    ...     mode=config.exposure,
    ...     detector=config.detector,
    ...     pipeline=config.pipeline,
    ...     with_intermediate_steps=True,
    ... )
    >>> results["/data/intermediate"]
    DataTree('intermediate', parent="data")
    │   Dimensions:  ()
    │   Data variables:
    │       *empty*
    │   Attributes:
    │       long_name:  Store all intermediate results modified along a pipeline
    └── DataTree('idx_0')
        │   Dimensions:  ()
        │   Data variables:
        │       *empty*
        │   Attributes:
        │       long_name:       Pipeline for one unique time
        │       pipeline_count:  0
        │       time:            1.0 s
        ├── DataTree('photon_collection')
        │   │   Dimensions:  ()
        │   │   Data variables:
        │   │       *empty*
        │   │   Attributes:
        │   │       long_name:  Model group: 'photon_collection'
        │   └── DataTree('load_image')
        │           Dimensions:  (y: 100, x: 100)
        │           Coordinates:
        │             * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        │             * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        │           Data variables:
        │               photon   (y, x) float64 1.515e+04 1.592e+04 ... 1.621e+04 1.621e+04
        │           Attributes:
        │               long_name:  Group: 'load_image'
        ├── DataTree('charge_generation')
        │   │   Dimensions:  ()
        │   │   Data variables:
        │   │       *empty*
        │   │   Attributes:
        │   │       long_name:  Model group: 'charge_generation'
        │   └── DataTree('photoelectrons')
        │           Dimensions:  (y: 100, x: 100)
        │           Coordinates:
        │             * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        │             * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        │           Data variables:
        │               charge   (y, x) float64 1.515e+04 1.592e+04 ... 1.621e+04 1.621e+04
        │           Attributes:
        │               long_name:  Group: 'photoelectrons'
        ├── DataTree('charge_collection')
        │   │   Dimensions:  ()
        │   │   Data variables:
        │   │       *empty*
        │   │   Attributes:
        │   │       long_name:  Model group: 'charge_collection'
        │   └── DataTree('simple_collection')
        │           Dimensions:  (y: 100, x: 100)
        │           Coordinates:
        │             * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        │             * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        │           Data variables:
        │               pixel    (y, x) float64 1.515e+04 1.592e+04 ... 1.621e+04 1.621e+04
        │           Attributes:
        │               long_name:  Group: 'simple_collection'
        ├── DataTree('charge_measurement')
        │   │   Dimensions:  ()
        │   │   Data variables:
        │   │       *empty*
        │   │   Attributes:
        │   │       long_name:  Model group: 'charge_measurement'
        │   └── DataTree('simple_measurement')
        │           Dimensions:  (y: 100, x: 100)
        │           Coordinates:
        │             * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        │             * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
        │           Data variables:
        │               signal   (y, x) float64 0.04545 0.04776 0.04634 ... 0.05004 0.04862 0.04862
        │           Attributes:
        │               long_name:  Group: 'simple_measurement'
        └── DataTree('readout_electronics')
            │   Dimensions:  ()
            │   Data variables:
            │       *empty*
            │   Attributes:
            │       long_name:  Model group: 'readout_electronics'
            └── DataTree('simple_adc')
                    Dimensions:  (y: 100, x: 100)
                    Coordinates:
                      * y        (y) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
                      * x        (x) int64 0 1 2 3 4 5 6 7 8 9 10 ... 90 91 92 93 94 95 96 97 98 99
                    Data variables:
                        image    (y, x) uint32 298 314 304 304 304 314 ... 325 339 339 328 319 319
                    Attributes:
                        long_name:  Group: 'simple_adc'
    """
    from pyxel.calibration import Calibration

    if with_intermediate_steps and isinstance(mode, (Observation, Calibration)):
        raise NotImplementedError(
            "Parameter 'with_intermediate_steps' is not implemented for 'Observation' and 'Calibration' modes."
        )

    if isinstance(mode, Exposure):
        data_tree = _run_exposure_mode(
            exposure=mode,
            detector=detector,
            pipeline=pipeline,
            with_intermediate_steps=with_intermediate_steps,
        )

    elif isinstance(mode, Observation):
        data_tree = _run_observation_mode(
            observation=mode,
            detector=detector,
            pipeline=pipeline,
        )

    elif isinstance(mode, Calibration):
        data_tree = _run_calibration_mode(
            calibration=mode,
            detector=detector,
            pipeline=pipeline,
        )

    else:
        raise TypeError("Please provide a valid simulation mode !")

    return data_tree


def output_directory(configuration: Configuration) -> Path:
    """Return the output directory from the configuration.

    Parameters
    ----------
    configuration

    Returns
    -------
    output_dir
    """
    # Late import to speedup start-up time
    from pyxel.calibration import Calibration

    if isinstance(configuration.exposure, Exposure):
        output_dir = configuration.exposure.outputs.output_dir
    elif isinstance(configuration.calibration, Calibration):
        output_dir = configuration.calibration.outputs.output_dir
    elif isinstance(configuration.observation, Observation):
        output_dir = configuration.observation.outputs.output_dir
    else:
        raise (TypeError("Outputs not initialized."))
    return output_dir


def run(input_filename: Union[str, Path], random_seed: Optional[int] = None) -> None:
    """Run a YAML configuration file.

    For more information, see :ref:`running_modes`.

    Parameters
    ----------
    input_filename : str or Path
    random_seed : int, optional

    Examples
    --------
    >>> import pyxel
    >>> pyxel.run("configuration.yaml")
    """
    # Late import to speedup start-up time

    logging.info("Pyxel version %s", version)
    logging.info("Pipeline started.")

    start_time = time.time()

    configuration: Configuration = load(Path(input_filename).expanduser().resolve())

    output_dir = output_directory(configuration)

    save(input_filename=input_filename, output_dir=output_dir)

    pipeline: DetectionPipeline = configuration.pipeline
    detector: Union[CCD, CMOS, MKID, APD] = configuration.detector
    running_mode: Union[Exposure, Observation, Calibration] = configuration.running_mode

    _ = run_mode(mode=running_mode, detector=detector, pipeline=pipeline)

    logging.info("Pipeline completed.")
    logging.info("Running time: %.3f seconds", (time.time() - start_time))
    # Closing the logger in order to be able to move the file in the output dir
    logging.shutdown()
    outputs.save_log_file(output_dir)

    # Late import to speedup start-up time
    from matplotlib import pyplot as plt

    plt.close()


# TODO: Add an option to display colors ?
@click.group()
@click.version_option(version=version)
def main():
    """Pyxel detector simulation framework.

    Pyxel is a detector simulation framework, that can simulate a variety of
    detector effects (e.g., cosmic rays, radiation-induced :term:`CTI` in :term:`CCDs<CCD>`, persistence
    in :term:`MCT`, charge diffusion, crosshatches, noises, crosstalk etc.) on a given image.
    """


@main.command(name="download-examples")
@click.argument("folder", type=click.Path(), default="pyxel-examples", required=False)
@click.option("-f", "--force", is_flag=True, help="Force flag for saving the examples.")
def download_pyxel_examples(folder, force: bool):
    """Install examples to a specified directory.

    Default folder is './pyxel-examples'.
    """
    download_examples(foldername=folder, force=force)


@main.command(name="create-model")
@click.argument("model_name", type=str)
def create_new_model(model_name: str):
    """Create a new model.

    Use: arg1/arg2. Create a new module in ``pyxel/models/arg1/arg2`` using a template
    (``pyxel/templates/MODELTEMPLATE.py``)
    """
    create_model(newmodel=model_name)


@main.command(name="run")
@click.argument("config", type=click.Path(exists=True))
@click.option(
    "-v",
    "--verbosity",
    count=True,
    show_default=True,
    help="Increase output verbosity (-v/-vv/-vvv)",
)
def run_config(config: str, verbosity: int):
    """Run Pyxel with a ``YAML`` configuration file."""
    logging_level = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG][
        min(verbosity, 3)
    ]
    log_format = (
        "%(asctime)s - %(name)s - %(threadName)30s - %(funcName)30s \t %(message)s"
    )
    logging.basicConfig(
        filename="pyxel.log",
        level=logging_level,
        format=log_format,
        datefmt="%d-%m-%Y %H:%M:%S",
    )

    # If user wants the log in stdout AND in file, use the three lines below
    stream_stdout = logging.StreamHandler(sys.stdout)
    stream_stdout.setFormatter(logging.Formatter(log_format))
    logging.getLogger().addHandler(stream_stdout)

    run(input_filename=config)


if __name__ == "__main__":
    main()



#
# pyxel documentation build configuration file, created by
# sphinx-quickstart on Tue Nov 21 09:37:42 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import sys
from datetime import datetime
from importlib.metadata import metadata

from sphinx.util import logging

import pyxel

# https://github.com/sphinx-doc/sphinx/issues/10112
logging.getLogger("sphinx.ext.extlinks").setLevel(40)

# Get metadata from 'pyproject.toml'
pyxel_metadata = metadata("pyxel-sim")

now_dt: datetime = datetime.now()

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",  # include documentation from docstrings
    "sphinx.ext.autosummary",  # Generate autodoc summaries
    "sphinx.ext.extlinks",
    "sphinx.ext.coverage",  # collect doc coverage stats
    "sphinx.ext.mathjax",  # render math via Javascript
    "sphinx.ext.napoleon",  # for numpy docstyle
    "sphinx.ext.todo",  # support for todo items (.. todo::)
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
    "sphinx.ext.viewcode",  # add links to highlighted source code
    "sphinxcontrib.bibtex",
    "sphinx_panels",
    "sphinx_inline_tabs",  # Add inline tabs
    "sphinx.ext.intersphinx",  # Link to other project's documentation
    "myst_nb",  # Parser '.ipynb' and '.md' files
    "sphinx_copybutton",  # Add a 'copy' button to code blocks
]

extlinks = {
    "issue": ("https://gitlab.com/esa/pyxel/-/issues/%s", "GH"),
    "pull": ("https://gitlab.com/esa/pyxel/-/merge_requests/%s", "PR"),
}
bibtex_bibfiles = ["refs.bib"]


autodoc_typehints = "none"

# Napoleon configurations
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_preprocess_types = True
napoleon_type_aliases = {
    # Objects related to a detector
    "Detector": "~pyxel.detectors.Detector",
    "CCD": "~pyxel.detectors.CCD",
    "CMOS": "~pyxel.detectors.CMOS",
    "MKID": "~pyxel.detectors.MKID",
    # Objects related to a running mode
    "Exposure": "~pyxel.exposure.Exposure",
    "Observation": "~pyxel.observation.Observation",
    "ObservationResult": "~pyxel.observation.ObservationResult",
    "Calibration": "~pyxel.calibration.Calibration",
    # Objects related to a pipeline
    "DetectionPipeline": "~pyxel.pipelines.DetectionPipeline",
    # General terms,
    "Sequence": ":term:`sequence`",
    "Path": ":py:class:`Path <pathlib.Path>`",
    # XArray
    "Dataset": "~xarray.Dataset",
    "DataTree": "~datatree.DataTree",
    # Pandas
    "DataFrame": "~pandas.DataFrame",
    # Numpy
    "ndarray": "~numpy.ndarray",
    # Holoviews
    "hv.Points": "~holoviews.element.Points",
    "hv.Layout": "~holoviews.core.Layout",
}

# For 'MyST-NB'
nb_execution_timeout = 240
suppress_warnings = ["mystnb.unknown_mime_type"]
myst_heading_anchors = 3

autodoc_member_order = "bysource"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = [".rst", ".md"]
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = pyxel_metadata["Name"]
copyright = f"2017-{now_dt:%Y}, European Space Agency"
author = "Pyxel Developers"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = pyxel.__version__
# The full version, including alpha/beta/rc tags.
release = pyxel.__version__
tag = version.split("+")[0]

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%Y-%m-%d"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
html_title = f"version {version}"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "repository_url": "https://gitlab.com/esa/pyxel",
    # Add buttons
    "use_edit_page_button": True,
    "use_repository_button": True,
    "use_issues_button": True,
    "use_download_button": True,
    "home_page_in_toc": False,
    "extra_footer": f"""<p>Last updated on {now_dt:%Y-%m-%d}.</p>""",
    "toc_title": "Contents",  # Control the right sidebar items
    # Version switcher dropdown. See issue #562
    # "switcher": {"json_url": "https://esa.gitlab.io/pyxel/doc/switcher.json"},
    # "check_switcher": False,
    # "announcement": "Hello World",
}

# This is used to generate the link 'suggest edit'
html_context = {
    "github_url": "https://gitlab.com",
    "github_user": "esa",
    "github_repo": "pyxel",
}
html_theme_options["path_to_docs"] = "../../-/edit/master/docs/source"


# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/pyxel-logo.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# configuration for sphinxext.opengraph
ogp_site_url = "https://esa.gitlab.io/pyxel/doc/"
ogp_image = "https://esa.gitlab.io/pyxel/doc/_static/esa-logo.png"
ogp_custom_meta_tags = [
    # '<meta name="twitter:card" content="summary_large_image" />',
    '<meta name="image" property="og:image" content="https://esa.gitlab.io/pyxel/doc/_static/esa-logo.png">',
]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
# html_sidebars = {
#     "**": [
#         "relations.html",  # needs 'show_related': True theme option to display
#         "searchbox.html",
#     ]
# }


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = pyxel_metadata["Name"] + "doc"


# https://xarray.pydata.org/en/stable/objects.inv
# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    # "iris": ("https://scitools-iris.readthedocs.io/en/latest", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    # "scipy": ("https://docs.scipy.org/doc/scipy", None),
    # "numba": ("https://numba.pydata.org/numba-doc/latest", None),
    # "matplotlib": ("https://matplotlib.org/stable/", None),
    "dask": ("https://docs.dask.org/en/latest", None),
    # "cftime": ("https://unidata.github.io/cftime", None),
    # "rasterio": ("https://rasterio.readthedocs.io/en/latest", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
    "holoviews": ("https://holoviews.org/", None),
    "datatree": ("https://xarray-datatree.readthedocs.io/en/latest/", None),
}


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        pyxel_metadata["Name"] + ".tex",
        pyxel_metadata["Name"] + " Documentation",
        pyxel_metadata["Author-email"],
        "manual",
    )
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        pyxel_metadata["Name"],
        pyxel_metadata["Name"] + " Documentation",
        [author],
        1,
    )
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        pyxel_metadata["Name"],
        pyxel_metadata["Name"] + " Documentation",
        author,
        pyxel_metadata["Name"],
        "One line description of project.",
        "Miscellaneous",
    )
]


def html_page_context(app, pagename, templatename, context, doctree):
    # Disable edit button for docstring generated pages
    if "generated" in pagename:
        context["theme_use_edit_page_button"] = False


def setup(app):
    app.connect("html-page-context", html_page_context)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.


"""TBW."""

import operator
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Protocol, Union

from pyxel.observation import ParameterMode
from pyxel.outputs import Outputs, ValidFormat, ValidName

if TYPE_CHECKING:
    from pyxel.observation import ObservationResult

    class SaveToFile(Protocol):
        """TBW."""

        def __call__(self, data: Any, name: str, with_auto_suffix: bool = True) -> Path:
            """TBW."""
            ...


class ObservationOutputs(Outputs):
    """Collection of methods to save the data buckets from a Detector for an Observation pipeline.

    Parameters
    ----------
    output_folder : str or Path
        Folder where sub-folder(s) that will be created to save data buckets.
    custom_dir_name : str, optional
        Prefix of the sub-folder name that will be created in the 'output_folder' folder.
        The default prefix is `run_`.
    save_data_to_file : Dict
        Dictionary where key is a 'data bucket' name (e.g. 'detector.photon.array') and value
        is the data format (e.g. 'fits').

        Example:
        {'detector.photon.array': 'fits', 'detector.charge.array': 'hdf', 'detector.image.array':'png'}
    """

    def __init__(
        self,
        output_folder: Union[str, Path],
        custom_dir_name: Optional[str] = "",
        save_data_to_file: Optional[
            Sequence[Mapping[ValidName, Sequence[ValidFormat]]]
        ] = None,
        save_observation_data: Optional[Sequence[Mapping[str, Sequence[str]]]] = None,
    ):
        super().__init__(
            output_folder=output_folder,
            save_data_to_file=save_data_to_file,
            custom_dir_name=custom_dir_name,
        )

        self.save_observation_data: Optional[
            Sequence[Mapping[str, Sequence[str]]]
        ] = save_observation_data

    # TODO: This function will be deprecated (see #563)
    def save_observation_datasets(
        self, result: "ObservationResult", mode: "ParameterMode"
    ) -> None:
        """Save the result datasets from parametric mode on disk.

        Parameters
        ----------
        result: Result
        mode: ParameterMode
        """

        dataset_names = ("dataset", "parameters", "logs")

        save_methods: dict[str, SaveToFile] = {"nc": self.save_to_netcdf}

        if self.save_observation_data is not None:
            dct: Mapping[str, Sequence[str]]
            for dct in self.save_observation_data:
                first_item, *_ = dct.items()
                obj, format_list = first_item

                if obj not in dataset_names:
                    raise ValueError(
                        "Please specify a valid result dataset names ('dataset', 'parameters', 'logs')."
                    )

                if mode == ParameterMode.Sequential and obj == "dataset":
                    dct = operator.attrgetter(obj)(result)
                    for key, value in dct.items():
                        if format_list is not None:
                            for out_format in format_list:
                                if out_format not in save_methods:
                                    raise ValueError(
                                        "Format "
                                        + out_format
                                        + " not a valid save method!"
                                    )

                                func = save_methods[out_format]
                                func(data=value, name=obj + "_" + key)

                else:
                    ds = operator.attrgetter(obj)(result)

                    if format_list is not None:
                        for out_format in format_list:
                            if out_format not in save_methods:
                                raise ValueError(
                                    "Format " + out_format + " not a valid save method!"
                                )

                            func = save_methods[out_format]
                            func(data=ds, name=obj)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Module to read/write ASDF files."""

from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    import pandas as pd
    import xarray as xr


def to_asdf(filename: Union[str, Path], dct: Mapping[str, Any]) -> None:
    """Write data to a ASDF file."""
    try:
        import asdf
    except ImportError as exc:
        raise ImportError(
            "Missing optional package 'asdf'.\n"
            "Please install it with 'pip install pyxel-sim[io]' "
            "or 'pip install pyxel-sim[all]'"
        ) from exc

    if dct["version"] != 1:
        raise NotImplementedError

    # Convert a 'DataFrame' into a `dict`
    df: pd.DataFrame = dct["data"]["charge"]["frame"]
    frame_dct: Mapping[str, Sequence[float]] = df.to_dict(orient="list")

    dct["data"]["charge"]["frame"] = frame_dct

    # Convert 'Dataset(s)' into a 'dict(s)'
    data: Mapping[str, xr.Dataset] = dct["data"]["data"]
    dct_data: Mapping[str, Mapping] = {
        key: value.to_dict() for key, value in data.items()
    }

    dct["data"]["data"] = dct_data

    with asdf.AsdfFile(dct) as af:
        af.write_to(filename)


@contextmanager
def from_asdf(filename: Union[str, Path]) -> Iterator[Mapping[str, Any]]:
    """Read data from a HDF5 file."""
    import pandas as pd

    try:
        import asdf
    except ImportError as exc:
        raise ImportError(
            "Missing optional package 'asdf'.\n"
            "Please install it with 'pip install pyxel-sim[io]' "
            "or 'pip install pyxel-sim[all]'"
        ) from exc

    dct: dict[str, Any] = {}

    with asdf.open(filename, copy_arrays=True) as af:
        # TODO: Use a JSON schema to validate 'dct'
        if "version" not in af:
            raise ValueError("Missing 'version' !")

        version: int = af["version"]
        if version != 1:
            raise NotImplementedError

        dct["version"] = version

        if "type" not in af:
            raise ValueError("Missing 'type' !")

        dct["type"] = af["type"]

        # Get properties
        dct["properties"] = af["properties"]
        dct["data"] = af["data"]

        # Convert a 'dict' to a 'DataFrame'
        frame_dct: Mapping[str, Sequence[float]] = dct["data"]["charge"]["frame"]
        df = pd.DataFrame(frame_dct)

        dct["data"]["charge"]["frame"] = df

        yield dct



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Tests for pulse_processing model."""

import pytest

from pyxel.detectors import (
    CCD,
    MKID,
    CCDGeometry,
    Characteristics,
    Environment,
    MKIDGeometry,
)
from pyxel.models.phasing import pulse_processing


@pytest.fixture
def mkid_5x5() -> MKID:
    """Create a valid MKID detector."""
    detector = MKID(
        geometry=MKIDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    return detector


@pytest.mark.parametrize(
    "wavelength, responsivity, scaling_factor",
    [
        pytest.param(
            1.0,
            1.0,
            1.0,
        ),
    ],
)
def test_pulse_processing_valid(
    mkid_5x5: MKID, wavelength: float, responsivity: float, scaling_factor: float
):
    """Test model 'pulse_processing' with valid inputs."""
    pulse_processing(
        detector=mkid_5x5,
        wavelength=wavelength,
        responsivity=responsivity,
        scaling_factor=scaling_factor,
    )


@pytest.mark.parametrize(
    "wavelength, responsivity, scaling_factor, exp_exc, exp_error",
    [
        pytest.param(
            -1.0,
            1.0,
            1.0,
            ValueError,
            "Only positive values accepted for wavelength.",
        ),
        pytest.param(
            1.0,
            1.0,
            -1.0,
            ValueError,
            "Only positive values accepted for scaling_factor.",
        ),
        pytest.param(
            1.0,
            -1.0,
            1.0,
            ValueError,
            "Only positive values accepted for responsivity.",
        ),
    ],
)
def test_pulse_processing_invalid(
    mkid_5x5: MKID,
    wavelength: float,
    responsivity: float,
    scaling_factor: float,
    exp_exc,
    exp_error,
):
    """Test model 'pulse_processing' with valid inputs."""
    with pytest.raises(exp_exc, match=exp_error):
        pulse_processing(
            detector=mkid_5x5,
            wavelength=wavelength,
            responsivity=responsivity,
            scaling_factor=scaling_factor,
        )


def test_pulse_processing_with_ccd():
    """Test model 'pulse_processing' with a `CCD` detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )

    with pytest.raises(TypeError, match="Expecting a MKID object for the detector."):
        pulse_processing(
            detector=detector, wavelength=1.0, responsivity=1.0, scaling_factor=1.0
        )



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel Scene class to track multi-wavelength photon."""

from collections.abc import Mapping
from typing import TYPE_CHECKING

import xarray as xr
from datatree import DataTree

if TYPE_CHECKING:
    from scopesim import Source


class Scene:
    """Scene class defining and storing information of all multi-wavelength photon."""

    def __init__(self):
        self._source: DataTree = DataTree(name="scene")

    def __eq__(self, other) -> bool:
        return type(self) is type(other) and self.data.identical(other.data)

    def add_source(self, source: xr.Dataset) -> None:
        """Add a source to the current scene.

        Parameters
        ----------
        source : Dataset

        Raises
        ------
        TypeError
            If 'source' is not a ``Dataset`` object.
        ValueError
            If 'source' has not the expected format.

        Examples
        --------
        >>> from pyxel.detectors import CCD
        >>> detector = CCD(...)
        >>> detector.reset()

        >>> source
        <xarray.Dataset>
        Dimensions:     (ref: 345, wavelength: 343)
        Coordinates:
          * ref         (ref) int64 0 1 2 3 4 5 6 7 ... 337 338 339 340 341 342 343 344
          * wavelength  (wavelength) float64 336.0 338.0 340.0 ... 1.018e+03 1.02e+03
        Data variables:
            x           (ref) float64 1.334e+03 1.434e+03 ... -1.271e+03 -1.381e+03
            y           (ref) float64 -1.009e+03 -956.1 -797.1 ... 1.195e+03 1.309e+03
            weight      (ref) float64 11.49 14.13 15.22 14.56 ... 15.21 11.51 8.727
            flux        (ref, wavelength) float64 0.003769 0.004137 ... 0.1813 0.1896

        >>> detector.scene.add_source(source)
        >>> detector.scene.data
        DataTree('scene', parent=None)
        └── DataTree('list')
            └── DataTree('0')
                    Dimensions:     (ref: 4, wavelength: 4)
                    Coordinates:
                      * ref         (ref) int64 0 1 2 3
                      * wavelength  (wavelength) float64 336.0 338.0 1.018e+03 1.02e+03
                    Data variables:
                        x           (ref) float64 64.97 11.44 -55.75 -20.66
                        y           (ref) float64 89.62 -129.3 -48.16 87.87
                        weight      (ref) float64 14.73 12.34 14.63 14.27
                        flux        (ref, wavelength) float64 0.003769 0.004137 ... 0.1813 0.1896
        """
        if not isinstance(source, xr.Dataset):
            raise TypeError("Expecting a Dataset object for source")

        if set(source.coords) != {"ref", "wavelength"}:
            raise ValueError(
                "Wrong format for source. Expecting coordinates 'ref' and 'wavelength'."
            )

        if set(source.data_vars) != {
            "x",
            "y",
            "weight",
            "flux",
        }:
            raise ValueError(
                "Wrong format for source. Expecting a Dataset with variables 'x', 'y', 'weight' and 'flux'."
            )

        if "list" not in self.data:
            key: int = 0
        else:
            key = self.data.width

        self.data[f"/list/{key}"] = DataTree(source)

    # TODO: This method will be removed in the future.
    #       If you want to have a `Source` object, you should use method '.to_scopesim'
    @property
    def data(self) -> DataTree:
        """Get a multi-wavelength object."""
        return self._source

    def from_scopesim(self, source: "Source") -> None:
        """Convert a ScopeSim `Source` object into a `Scene` object.

        Parameters
        ----------
        source : scopesim.Source
            Object to convert to a `Scene` object.

        Raises
        ------
        RuntimeError
            If package 'scopesim' is not installed.
        TypeError
            If input parameter 'source' is not a ScopeSim `Source` object.

        Notes
        -----
        More information about ScopeSim `Source` objects at
        this link: https://scopesim.readthedocs.io/en/latest/reference/scopesim.source.source.html
        """
        try:
            from scopesim import Source
        except ImportError as exc:
            raise RuntimeError(
                "Package 'scopesim' is not installed ! "
                "Please run command 'pip install scopesim' from the command line."
            ) from exc

        if not isinstance(source, Source):
            raise TypeError("Expecting a ScopeSim `Source` object for 'source'.")

        raise NotImplementedError

    def to_scopesim(self) -> "Source":
        """Convert this `Scene` object into a ScopeSim `Source` object.

        Returns
        -------
        Source
            A ScopeSim `Source` object.

        Notes
        -----
        More information about ScopeSim `Source` objects at
        this link: https://scopesim.readthedocs.io/en/latest/reference/scopesim.source.source.html
        """
        raise NotImplementedError

    def to_dict(self) -> Mapping:
        """Convert an instance of `Scene` to a `dict`."""
        result: Mapping = {
            key: value.to_dict() for key, value in self.data.to_dict().items()
        }

        return result

    @classmethod
    def from_dict(cls, dct: Mapping) -> "Scene":
        """Create a new instance of a `Scene` object from a `dict`."""
        data: Mapping[str, xr.Dataset] = {
            key: xr.Dataset.from_dict(value) for key, value in dct.items()
        }

        scene = cls()
        scene._source = DataTree.from_dict(data, name="scene")  # type: ignore

        return scene

    def to_xarray(self) -> xr.Dataset:
        """Convert current scene to a xarray Dataset.

        Returns
        -------
        xr.Dataset

        Examples
        --------
        >>> ds = scene.to_xarray()
        >>> ds
        <xarray.Dataset>
        Dimensions:     (ref: 345, wavelength: 343)
        Coordinates:
          * ref         (ref) int64 0 1 2 3 4 5 6 7 ... 337 338 339 340 341 342 343 344
          * wavelength  (wavelength) float64 336.0 338.0 340.0 ... 1.018e+03 1.02e+03
        Data variables:
            x           (ref) float64 2.057e+05 2.058e+05 ... 2.031e+05 2.03e+05
            y           (ref) float64 8.575e+04 8.58e+04 ... 8.795e+04 8.807e+04
            weight      (ref) float64 11.49 14.13 15.22 14.56 ... 15.21 11.51 8.727
            flux        (ref, wavelength) float64 0.03769 0.04137 ... 1.813 1.896
        >>> ds["wavelength"]
        <xarray.DataArray 'wavelength' (wavelength: 343)>
        array([ 336.,  338.,  340., ..., 1016., 1018., 1020.])
        Coordinates:
          * wavelength  (wavelength) float64 336.0 338.0 340.0 ... 1.018e+03 1.02e+03
        Attributes:
            units:    nm
        >>> ds["flux"]
        <xarray.DataArray 'flux' (ref: 345, wavelength: 343)>
        array([[3.76907117e-02, 4.13740861e-02, ..., 3.98815404e-02, 7.96581117e-01],
               [1.15190254e-02, 1.02210366e-02, ..., 2.00486326e-02, 2.05518196e-02],
               ...,
               [1.01187592e-01, 9.57637374e-02, ..., 2.71410354e-01, 2.85997559e-01],
               [1.80093381e+00, 1.69864354e+00, ..., 1.81295134e+00, 1.89642359e+00]])
        Coordinates:
          * ref         (ref) int64 0 1 2 3 4 5 6 7 ... 337 338 339 340 341 342 343 344
          * wavelength  (wavelength) float64 336.0 338.0 340.0 ... 1.018e+03 1.02e+03
        Attributes:
            units:    ph / (cm2 nm s)
        """
        assert len(self.data["/list"]) == 1
        data: xr.Dataset = self.data["/list/0"].to_dataset()

        return data



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from pathlib import Path

import numpy as np
import pytest
from datatree import DataTree

import pyxel
from pyxel import Configuration
from pyxel.detectors import Detector
from pyxel.exposure import Exposure


@pytest.fixture
def valid_config_filename(request: pytest.FixtureRequest) -> Path:
    """Get a valid existing YAML filename."""
    filename: Path = request.path.parent / "data/basic_exposure.yaml"
    return filename.resolve(strict=True)


def test_basic_exposure_hdf5(valid_config_filename: Path, tmp_path: Path):
    """Functional test with a basic Exposure mode."""
    # Read configuration file
    cfg = pyxel.load(valid_config_filename)
    assert isinstance(cfg, Configuration)

    # Save 'detector' before modifications
    detector = cfg.detector
    assert isinstance(detector, Detector)

    detector_filename_before = tmp_path / "before_detector.hdf5"
    assert not detector_filename_before.exists()

    detector.to_hdf5(detector_filename_before)
    assert detector_filename_before.exists()

    # Execute 'cfg'
    data_tree = pyxel.run_mode(
        mode=cfg.running_mode,
        detector=detector,
        pipeline=cfg.pipeline,
    )
    assert isinstance(data_tree, DataTree)

    # Save the 'detector' object into a '.hdf5' file
    detector_filename: Path = tmp_path / "detector.hdf5"
    assert not detector_filename.exists()

    detector.to_hdf5(detector_filename)
    assert detector_filename.exists()

    # Load to a new 'detector' object from '.hdf5' file
    new_detector = Detector.from_hdf5(detector_filename)

    assert detector.data.isomorphic(new_detector.data)
    assert set(detector.data.groups) == set(new_detector.data.groups)
    # assert detector == new_detector


def test_basic_exposure_asdf(valid_config_filename: Path, tmp_path: Path):
    """Functional test with a basic Exposure mode."""
    # Read configuration file
    cfg = pyxel.load(valid_config_filename)
    assert isinstance(cfg, Configuration)

    # Save 'detector' before modifications
    detector = cfg.detector
    assert isinstance(detector, Detector)

    detector_filename_before = tmp_path / "before_detector.asdf"
    assert not detector_filename_before.exists()

    detector.to_asdf(detector_filename_before)
    assert detector_filename_before.exists()

    # Execute 'cfg'
    data_tree = pyxel.run_mode(
        mode=cfg.running_mode,
        detector=cfg.detector,
        pipeline=cfg.pipeline,
    )
    assert isinstance(data_tree, DataTree)

    # Save the 'detector' object into a '.asdf' file
    detector_filename: Path = tmp_path / "detector.asdf"
    assert not detector_filename.exists()

    detector.to_asdf(detector_filename)
    assert detector_filename.exists()

    # Load to a new 'detector' object from '.asdf' file
    new_detector = Detector.from_asdf(detector_filename)

    assert detector.data.isomorphic(new_detector.data)
    assert set(detector.data.groups) == set(new_detector.data.groups)
    # assert detector == new_detector


@pytest.mark.parametrize("readout_times", [[0.1], [0.1, 0.5]])
def test_simple_adc_dtype(readout_times, valid_config_filename: Path):
    """Functional tests to check '.dtype' when using model 'simple_adc'."""
    # Read configuration file
    cfg = pyxel.load(valid_config_filename)
    assert isinstance(cfg, Configuration)

    detector = cfg.detector

    exposure = cfg.running_mode
    assert isinstance(exposure, Exposure)

    exposure.readout.times = readout_times

    result = pyxel.run_mode(
        mode=exposure,
        detector=detector,
        pipeline=cfg.pipeline,
    )

    assert detector.image.dtype == np.uint16
    assert result["image"].dtype == np.uint16



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Pyxel CosmiX model to generate charge by ionization."""

from typing import Literal, Optional, Union

import numpy as np

from pyxel.detectors import Detector
from pyxel.models.charge_generation.cosmix.util import sampling_distribution


class Particle:
    """Particle class define a particle together with its characteristics."""

    def __init__(
        self,
        detector: Detector,
        simulation_mode: Literal[
            "cosmic_ray", "cosmics", "radioactive_decay", "snowflakes"
        ],
        particle_type: Literal[
            "proton", "ion", "alpha", "beta", "electron", "gamma", "x-ray"
        ],
        input_energy: Union[int, float, Literal["random"]],
        spectrum_cdf: np.ndarray,
        starting_pos_ver: Union[str, np.ndarray],
        starting_pos_hor: Union[str, np.ndarray],
        starting_pos_z: Union[str, np.ndarray],
        # input_alpha='random', input_beta='random'
    ):
        """Creation of a particle according to some parameters.

        Parameters
        ----------
        detector
        simulation_mode
        particle_type
        input_energy
        spectrum_cdf
        starting_pos_ver
        starting_pos_hor
        starting_pos_z
        """
        self.detector = detector
        geo = self.detector.geometry

        # starting_position_vertical = None
        # starting_position_horizontal = None
        # starting_position_z = None

        # if starting_pos_ver == 'random':
        #     starting_position_vertical = geo.vert_dimension * np.random.random()
        # elif isinstance(starting_pos_ver, int) or isinstance(starting_pos_ver, float):
        #     starting_position_vertical = starting_pos_ver
        # if starting_pos_hor == 'random':
        #     starting_position_horizontal = geo.horz_dimension * np.random.random()
        # elif isinstance(starting_pos_hor, int) or isinstance(starting_pos_hor, float):
        #     starting_position_horizontal = starting_pos_hor
        #
        # if starting_pos_z == 'random':
        #     starting_position_z = geo.total_thickness * np.random.random()
        # elif isinstance(starting_pos_z, int) or isinstance(starting_pos_z, float):
        #     starting_position_z = starting_pos_z
        #
        # self.starting_position = np.array([starting_position_vertical,
        #                                    starting_position_horizontal,
        #                                    starting_position_z])
        # self.position = np.copy(self.starting_position)
        # self.trajectory = np.copy(self.starting_position)

        # if input_alpha == 'random' and starting_pos_z == 0.:
        #     alpha = 2 * math.pi * np.random.random()
        # elif input_alpha == 'random' and starting_pos_z != 0.:
        #     alpha = 2 * math.pi * np.random.random()
        # else:
        #     alpha = input_alpha  # between 0 and 2*pi
        #
        # if input_beta == 'random':
        #     beta = 2. * math.pi * np.random.random()
        # else:
        #     beta = input_beta
        # self.angle = np.array([alpha, beta])
        #
        # self.dir_z = -1 * math.sin(alpha)
        # self.dir_ver = math.cos(alpha) * math.cos(beta)
        # self.dir_hor = math.cos(alpha) * math.sin(beta)

        # if input_alpha != 'random':
        #     self.alpha = input_alpha
        # if input_beta != 'random':
        #     self.beta = input_beta
        # # update direction:
        # self.dir_ver, self.dir_hor, self.dir_z = get_direction_from_angles()

        self.track_length: Optional[float] = None

        self.dir_ver, self.dir_hor, self.dir_z = isotropic_direction()

        self.random_det_pt_vert = geo.vert_dimension * np.random.random()
        self.random_det_pt_horz = geo.horz_dimension * np.random.random()
        self.random_det_pt_z = -1 * geo.total_thickness * np.random.random()

        self.alpha, self.beta = self.get_angles()  # rad

        mode_1 = ["cosmic_ray", "cosmics"]
        mode_2 = ["radioactive_decay", "snowflakes"]
        if (
            simulation_mode in mode_1
        ):  # cosmic rays coming from OUTSIDE the detector volume
            self.starting_position = self.get_surface_point()
        elif simulation_mode in mode_2:  # radioactive decay INSIDE the detector volume
            self.starting_position = np.array(
                [self.random_det_pt_vert, self.random_det_pt_horz, self.random_det_pt_z]
            )

        if starting_pos_ver != "random":
            self.starting_position[0] = starting_pos_ver
        if starting_pos_hor != "random":
            self.starting_position[1] = starting_pos_hor
        if starting_pos_z != "random":
            self.starting_position[2] = starting_pos_z

        self.position = np.copy(self.starting_position)
        self.trajectory = np.copy(self.starting_position)

        if input_energy == "random":
            self.energy: float = sampling_distribution(spectrum_cdf)
        elif isinstance(input_energy, (int, float)):
            self.energy = input_energy
        else:
            raise ValueError("Given particle energy could not be read")

        self.deposited_energy: float = 0.0
        self.total_edep: float = 0.0

        self.type = particle_type
        ionizing_particles = ("proton", "ion", "alpha", "beta", "electron")
        non_ionizing_particles = ("gamma", "x-ray")  # 'photon'

        if self.type in ionizing_particles:
            # call direct ionization func when needed - already implemented in simulation
            pass

        elif self.type in non_ionizing_particles:
            # call NON-direct ionization func when needed - need to be implemented
            raise NotImplementedError(
                "Given particle type simulation is not yet implemented"
            )

        else:
            raise ValueError("Given particle type can not be simulated")

    def get_surface_point(self) -> np.ndarray:
        geo = self.detector.geometry

        norm_vectors: list[np.ndarray] = [
            np.array(
                [0.0, 0.0, -1.0]
            ),  # top plane (usually particle enters vol. via this)
            np.array(
                [0.0, 0.0, 1.0]
            ),  # bottom plane (usually particle leaves vol. via this)
            np.array([0.0, 1.0, 0.0]),
            np.array([-1.0, 0.0, 0.0]),
            np.array([0.0, -1.0, 0.0]),
            np.array([1.0, 0.0, 0.0]),
        ]

        points: list[np.ndarray] = [
            np.array(
                [0.0, 0.0, 0.0]
            ),  # top plane (usually particle enters vol. via this)
            np.array(
                [0.0, 0.0, -1 * geo.total_thickness]
            ),  # bottom plane (usually particle leaves vol. via this)
            np.array([0.0, 0.0, 0.0]),
            np.array([geo.vert_dimension, 0.0, 0.0]),
            np.array([geo.vert_dimension, geo.horz_dimension, 0.0]),
            np.array([0.0, geo.horz_dimension, 0.0]),
        ]

        intersect_points = np.zeros((6, 3))
        track_direction = np.array([self.dir_ver, self.dir_hor, self.dir_z])
        random_det_point = np.array(
            [self.random_det_pt_vert, self.random_det_pt_horz, self.random_det_pt_z]
        )

        surface_start_point: Optional[np.ndarray] = None  #
        surface_end_point: Optional[np.ndarray] = None
        for i in range(6):
            intersect_points[i, :] = find_intersection(
                n=norm_vectors[i], p0=points[i], ls=random_det_point, lv=track_direction
            )

            eps = 1e-8
            if (
                (0.0 - eps <= intersect_points[i, 0] <= geo.vert_dimension + eps)
                and (0.0 - eps <= intersect_points[i, 1] <= geo.horz_dimension + eps)
                and (
                    -1 * geo.total_thickness - eps
                    <= intersect_points[i, 2]
                    <= 0.0 + eps
                )
            ):
                if (
                    np.dot(
                        track_direction,
                        intersect_points[i, :] - random_det_point,
                    )
                    < 0
                ):
                    surface_start_point = self.intersection_correction(
                        intersect_points[i, :]
                    )
                else:
                    surface_end_point = self.intersection_correction(
                        intersect_points[i, :]
                    )

        if surface_start_point is None or surface_end_point is None:
            raise RuntimeError(
                "'surface_start_point' and/or 'surface_end_point' are not defined."
            )

        surface_1d: np.ndarray = surface_end_point - surface_start_point
        self.track_length = float(np.linalg.norm(x=surface_1d))

        if np.all(surface_start_point == surface_end_point):
            raise ValueError("This should not happen")
        if surface_start_point is None:
            raise ValueError("This should not happen")
        if surface_end_point is None:
            raise ValueError("This should not happen")

        return surface_start_point

    def get_angles(self) -> tuple[float, float]:
        beta = np.arccos(
            np.dot(np.array([1.0, 0.0]), np.array([self.dir_ver, self.dir_hor]))
        )

        alpha = np.arccos(
            np.dot(
                np.array([0.0, 0.0, 1.0]),
                np.array([self.dir_ver, self.dir_hor, self.dir_z]),
            )
        )

        if self.dir_hor < 0.0:
            beta += np.pi
            alpha += np.pi

        return alpha, beta

    # def track_length(self):
    #     """TBW.
    #
    #     :return:
    #     """
    #     geo = self.detector.geometry
    #
    #     norm_vectors = [np.array([0., 0., -1.]),    # top plane (usually particle enters vol. via this)
    #                     np.array([0., 0., 1.]),     # bottom plane (usually particle leaves vol. via this)
    #                     np.array([0., 1., 0.]),
    #                     np.array([-1., 0., 0.]),
    #                     np.array([0., -1., 0.]),
    #                     np.array([1., 0., 0.])]
    #
    #     points = [np.array([0., 0., 0.]),                       # top plane (usually particle enters vol. via this)
    #               np.array([0., 0., -1 * geo.total_thickness]), # bottom plane (usually particle leaves vol. via this)
    #               np.array([0., 0., 0.]),
    #               np.array([geo.vert_dimension, 0., 0.]),
    #               np.array([geo.vert_dimension, geo.horz_dimension, 0.]),
    #               np.array([0., geo.horz_dimension, 0.])]
    #
    #     track_length = np.inf
    #     intersect_points = np.zeros((6, 3))
    #     dir_array = np.array([self.dir_ver,
    #                           self.dir_hor,
    #                           self.dir_z])
    #     for i in range(6):
    #         intersect_points[i, :] = find_intersection(n=norm_vectors[i], p0=points[i],
    #                                                    ls=self.starting_position, lv=dir_array)
    #         track_length_new = np.linalg.norm(intersect_points[i, :] - self.starting_position)
    #         if track_length_new < track_length and track_length_new != 0.:
    #             track_length = track_length_new
    #
    #     return track_length

    # TODO: warning this method modify input parameter 'array' !!!
    def intersection_correction(self, array: np.ndarray) -> np.ndarray:
        eps = 1e-8
        geo = self.detector.geometry
        if abs(array[0] - geo.vert_dimension) < eps:
            array[0] = geo.vert_dimension
        if abs(array[0]) < eps:
            array[0] = 0.0
        if abs(array[1] - geo.horz_dimension) < eps:
            array[1] = geo.horz_dimension
        if abs(array[1]) < eps:
            array[1] = 0.0
        if abs(array[2] + geo.total_thickness) < eps:
            array[2] = -1 * geo.total_thickness
        if abs(array[2]) < eps:
            array[2] = 0.0

        return array


def find_intersection(
    n: np.ndarray, p0: np.ndarray, ls: np.ndarray, lv: np.ndarray
) -> Optional[np.ndarray]:
    """TBW.

    https://en.wikipedia.org/wiki/Line%E2%80%93plane_intersection
    :param n: normal vector of the plane
    :param p0: point of the plane
    :param ls: starting point of particle track
    :param lv: direction of particle track
    :return:
    """
    if np.dot(lv, n) == 0:  # No intersection of track and detector plane
        return None
    else:
        d = np.dot((p0 - ls), n) / np.dot(lv, n)
        p: np.ndarray = d * lv + ls
        return p


def isotropic_direction() -> tuple[float, float, float]:
    u = 2 * np.random.random() - 1
    r = np.sqrt(1 - u**2)
    kszi = np.random.random()
    v = r * np.cos(2 * np.pi * kszi)
    w = r * np.sin(2 * np.pi * kszi)
    return u, v, w


def non_isotropic_direction(n: int) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    alpha = 2 * np.pi * np.random.random(n)
    beta = 2 * np.pi * np.random.random(n)
    x = np.cos(alpha) * np.sin(beta)
    y = np.cos(alpha) * np.cos(beta)
    z = np.sin(alpha)
    return x, y, z



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

from pathlib import Path

import numpy as np
import pytest
import xarray as xr
from astropy.io import fits

from pyxel.calibration import create_processor_data_array


@pytest.fixture
def folder_files(tmp_path: Path):
    """Create dummy files."""

    np.save(file=tmp_path / "data00.npy", arr=np.zeros(shape=(10, 2)))
    np.save(file=tmp_path / "data01.npy", arr=np.full(shape=(10, 2), fill_value=1.0))
    np.save(file=tmp_path / "data02.npy", arr=np.full(shape=(10, 2), fill_value=2.0))
    np.save(file=tmp_path / "data03.npy", arr=np.full(shape=(10, 2), fill_value=3.0))
    fits.writeto(
        filename=tmp_path / "data04.fits", data=np.full(shape=(10, 2), fill_value=4.0)
    )

    return tmp_path


def test_create_processor_data_array(folder_files: Path):
    """Test function 'create_processor_data_array'."""
    filenames = [
        folder_files / "data00.npy",
        folder_files / "data01.npy",
        folder_files / "data02.npy",
        folder_files / "data03.npy",
        folder_files / "data04.fits",
    ]

    result = create_processor_data_array(filenames)
    exp_result = xr.DataArray(
        np.broadcast_to(
            np.array([0, 1, 2, 3, 4], dtype=float).reshape((5, 1, 1)), shape=(5, 10, 2)
        ),
        dims=["processor", "y", "x"],
        coords={"processor": range(5), "y": range(10), "x": range(2)},
    )

    xr.testing.assert_identical(result, exp_result)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Tests for dark current models."""

import pytest

from pyxel.detectors import (
    APD,
    CCD,
    APDCharacteristics,
    APDGeometry,
    CCDGeometry,
    Characteristics,
    Environment,
    ReadoutProperties,
)
from pyxel.models.charge_generation import (
    dark_current,
    dark_current_saphira,
    simple_dark_current,
)


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(temperature=200.0),
        characteristics=Characteristics(),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


@pytest.fixture
def apd_5x5() -> APD:
    """Create a valid CCD detector."""
    detector = APD(
        geometry=APDGeometry(
            row=5,
            col=5,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(temperature=50.0),
        characteristics=APDCharacteristics(
            quantum_efficiency=1.0,
            adc_voltage_range=(0.0, 10.0),
            adc_bit_resolution=16,
            full_well_capacity=100000,
            avalanche_gain=10.0,
            pixel_reset_voltage=5.0,
            roic_gain=0.8,
        ),
    )
    detector._readout_properties = ReadoutProperties(times=[1.0])
    return detector


def test_simple_dark_current_valid(ccd_10x10: CCD):
    """Test model 'simple_dark_current' with valid inputs."""
    simple_dark_current(detector=ccd_10x10, dark_rate=1.0)


def test_dark_current_valid(ccd_10x10: CCD):
    """Test model 'dark_current' with valid inputs."""
    dark_current(detector=ccd_10x10, figure_of_merit=1.0, spatial_noise_factor=0.4)


def test_dark_current_warning(ccd_10x10: CCD):
    """Test model 'dark_current' when generating a warning."""
    detector = ccd_10x10
    detector.environment.temperature = 300

    with pytest.warns(RuntimeWarning, match="Unphysical high value"):
        dark_current(detector=detector, figure_of_merit=100.0, spatial_noise_factor=0.4)


@pytest.mark.parametrize(
    "figure_of_merit, spatial_noise_factor, band_gap, band_gap_room_temperature, exp_exc, exp_error",
    [
        pytest.param(
            1.0,
            0.01,
            None,
            1.2,
            ValueError,
            "Both parameters band_gap and band_gap_room_temperature have to be defined.",
        ),
        pytest.param(
            1.0,
            0.01,
            1.2,
            None,
            ValueError,
            "Both parameters band_gap and band_gap_room_temperature have to be defined.",
        ),
    ],
)
def test_dark_current_invalid(
    ccd_10x10: CCD,
    figure_of_merit: float,
    spatial_noise_factor: float,
    band_gap: float,
    band_gap_room_temperature: float,
    exp_exc,
    exp_error,
):
    """Test model 'dark_current' with valid inputs."""
    with pytest.raises(exp_exc, match=exp_error):
        dark_current(
            detector=ccd_10x10,
            figure_of_merit=figure_of_merit,
            spatial_noise_factor=spatial_noise_factor,
            band_gap=band_gap,
            band_gap_room_temperature=band_gap_room_temperature,
        )


def test_dark_current_saphira_valid(apd_5x5: APD):
    """Test model 'dark_current_saphira' with valid inputs."""
    dark_current_saphira(detector=apd_5x5)


def test_dark_current_saphira_with_ccd(ccd_10x10: CCD):
    """Test model 'dark_current_saphira' with a 'CCD'."""
    detector = ccd_10x10

    with pytest.raises(TypeError, match="Expecting an APD object for detector."):
        dark_current_saphira(detector=detector)


@pytest.mark.parametrize(
    "temperature, gain, exp_exc, exp_error",
    [
        pytest.param(
            10.0,
            1.0,
            ValueError,
            "Dark current is inaccurate for avalanche gains less than 2!",
        ),
        pytest.param(
            200.0,
            10.0,
            ValueError,
            "Dark current estimation is inaccurate for temperatures more than 100 K!",
        ),
    ],
)
def test_dark_current_saphira_invalid(
    apd_5x5: APD, temperature: float, gain: float, exp_exc, exp_error
):
    """Test model 'dark_current_saphira' with valid inputs."""
    detector = apd_5x5
    with pytest.raises(exp_exc, match=exp_error):  # noqa
        detector.environment.temperature = temperature
        detector.characteristics.avalanche_gain = gain
        dark_current_saphira(detector=detector)



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

"""Scene generator creates Scopesim Source object."""

from collections.abc import Sequence
from enum import Enum
from typing import TYPE_CHECKING, Literal

import astropy.units as u
import numpy as np
import requests
import xarray as xr
from astroquery.gaia import Gaia
from specutils import Spectrum1D
from synphot import SourceSpectrum

from pyxel.detectors import Detector

if TYPE_CHECKING:
    from astropy.io.votable import tree
    from astropy.table import Table


class GaiaPassBand(Enum):
    """Define different type of magnitude provided by the Gaia database.

    More information available here: https://www.cosmos.esa.int/web/gaia/iow_20180316
    """

    BluePhotometer = "blue_photometer"  # Wavelength from 330 nm to 680 nm
    GaiaBand = "gaia_band"  # Wavelength from 330 nm to 1050 nm
    RedPhotometer = "red_photometer"  # Wavelength from 640 nm to 1050 nm

    def get_magnitude_key(self) -> str:
        """Return the Gaia magnitude keyword.

        Examples
        --------
        >>> band = GaiaPassBand.BluePhotometer
        >>> band.get_magnitude_key()
        'phot_bp_mean_map'
        """
        if self is self.BluePhotometer:
            return "phot_bp_mean_mag"
        elif self is self.GaiaBand:
            return "phot_g_mean_mag"
        elif self is self.RedPhotometer:
            return "phot_rp_mean_mag"
        else:
            raise NotImplementedError


def compute_flux(wavelength: u.Quantity, flux: u.Quantity) -> u.Quantity:
    """Compute flux in Photon.

    Parameters
    ----------
    wavelength : Quantity
        Unit: nm
    flux : Quantity
        Unit: W / (nm * m^2)

    Returns
    -------
    Quantity
        Unit: photon / (s * cm^2 * nm)

    >>> wavelength
    <Quantity [ 336.,  338.,  340., ..., 1016., 1018., 1020.] nm>
    >>> flux
    <Quantity [1.0647195e-14, 9.9830278e-15, 9.1758579e-15, ..., 3.5089168e-15,
               3.5376517e-15, 3.6932771e-15] W / (nm m2)>
    >>> compute_flux(wavelength=wavelength, flux=flux)
    <Quantity [0.18009338, 0.18009338, 0.18009338, ..., 0.18009338, 0.18009338,
               0.18009338] ph / (nm s cm2)>
    """
    spectrum_1d = Spectrum1D(spectral_axis=wavelength, flux=flux)
    source_spectrum = SourceSpectrum.from_spectrum1d(spectrum_1d)

    flux_photlam: u.Quantity = source_spectrum(wavelength)

    return flux_photlam.to(u.ph / (u.s * u.cm * u.cm * u.nm))


def _convert_spectrum(flux_1d: np.ndarray, wavelength_1d: np.ndarray) -> np.ndarray:
    """Convert a 1d flux in W / (nm m2) to ph / (s AA cm2).

    Parameters
    ----------
    flux_1d : np.ndarray
        1D numpy array
    wavelength_1d : np.ndarray
        1D numpy array

    Returns
    -------
    np.ndarray

    Examples
    --------
    >>> import numpy as np
    >>> flux_1d = np.array(
    ...     [2.2282904e-16, 2.4315793e-16, ..., 1.5625901e-15, 1.6213707e-15],
    ...     dtype=np.float32,
    ... )
    >>> wavelength_1d = np.array(
    ...     [3.76907117e-03, 4.13740861e-03, ..., 8.00785350e-02, 8.32541239e-02]
    ... )
    >>> _convert_spectrum(flux_1d=flux_1d, wavelength_1d=wavelength_1d)
    array([3.76907117e-03, 4.13740861e-03, ..., 8.00785350e-02, 8.32541239e-02])
    """
    converted_flux: u.Quantity = compute_flux(
        wavelength=wavelength_1d * u.nm,
        flux=flux_1d * (u.W / (u.nm * u.m * u.m)),
    )

    return np.asarray(converted_flux)


def convert_spectrum(flux: xr.DataArray) -> xr.DataArray:
    """Convert a DataArray flux in W / (nm m2) to ph / (s AA cm2).

    Parameters
    ----------
    flux : DataArray
        This DataArray must have at least one dimension 'wavelength'

    Returns
    -------
    DataArray

    Examples
    --------
    >>> import xarray as xr
    >>> flux = xr.DataArray(...)
    >>> flux
    <xarray.DataArray 'flux' (source_id: 345, wavelength: 343)>
    array([[2.2282904e-16, 2.4315793e-16, ..., 1.5625901e-15, 1.6213707e-15],
           [6.8100953e-17, 6.0069631e-17, ..., 3.9121338e-17, 4.0024586e-17],
           ...,
           [5.9822517e-16, 5.6280908e-16, ..., 5.2960899e-16, 5.5697906e-16],
           [1.0647195e-14, 9.9830278e-15, ..., 3.5376517e-15, 3.6932771e-15]], dtype=float32)
    Coordinates:
      * source_id   (source_id) int64 66504343062472704 ... 65296357738731008
      * wavelength  (wavelength) float64 336.0 338.0 340.0 ... 1.018e+03 1.02e+03
    Attributes:
        units:    W / (nm m2)

    >>> convert_spectrum(flux)
    <xarray.DataArray 'flux' (source_id: 345, wavelength: 343)>
    array([[3.76907117e-02, 4.13740861e-02, ..., 8.00785350e-01, 8.32541239e-01],
           [1.15190254e-02, 1.02210366e-02, ..., 2.00486326e-02, 2.05518196e-02],
           ...,
           [1.01187592e-01, 9.57637374e-02, ..., 2.71410354e-01, 2.85997559e-01],
           [1.80093381e+00, 1.69864354e+00, ..., 1.81295134e+00, 1.89642359e+00]])
    Coordinates:
      * source_id   (source_id) int64 66504343062472704 ... 65296357738731008
      * wavelength  (wavelength) float64 336.0 338.0 340.0 ... 1.018e+03 1.02e+03
    Attributes:
        units:    ph / (nm s cm2)
    """
    new_flux: xr.DataArray = xr.apply_ufunc(
        _convert_spectrum,
        flux,
        kwargs={"wavelength_1d": flux["wavelength"].to_numpy()},
        vectorize=True,
        input_core_dims=[["wavelength"]],
        output_core_dims=[["wavelength"]],
    )

    new_flux.attrs["units"] = str(u.ph / (u.s * u.nm * u.cm * u.cm))
    new_flux["wavelength"].attrs["units"] = flux["wavelength"].units
    return new_flux


def _retrieve_objects_from_gaia(
    right_ascension: float,
    declination: float,
    fov_radius: float,
) -> tuple["Table", dict[int, "Table"]]:
    """Retrieve objects from GAIA Catalog for giver coordinates and FOV.

    Columns description:
    * ``source_id``: Unique source identifier of the source
    * ``ra``: Barycentric right ascension of the source.
    * ``dec``: Barycentric Declination of the source.
    * ``has_xp_sampled``: Flag indicating the availability of mean BP/RP spectrum in sampled form for this source
    * ``phot_bp_mean_mag``: Mean magnitude in the integrated BP band.
    * ``phot_g_mean_mag``: Mean magnitude in the G band.
    * ``phot_rp_mean_mag``: Mean magnitude in the integrated RP band.

    Parameters
    ----------
    right_ascension: float
        RA coordinate in degree.
    declination: float
        DEC coordinate in degree.
    fov_radius: float
        FOV radius of telescope optics.

    Returns
    -------
    Table, Sequence of Tables
        All sources found and a list of spectrum for these sources

    Raises
    ------
    ConnectionError
        If the connection to the GAIA database cannot be established.

    Notes
    -----
    More information about the GAIA catalog at these links:
    * https://gea.esac.esa.int/archive/documentation/GDR3/
    * https://gea.esac.esa.int/archive/documentation/GDR3/Gaia_archive/chap_datamodel/sec_dm_main_source_catalogue/ssec_dm_gaia_source.html

    Examples
    --------
    >>> positions, spectra = _retrieve_objects_from_gaia(
    ...     right_ascension=56.75,
    ...     declination=24.1167,
    ...     fov_radius=0.05,
    ... )
    >>> positions
        source_id             ra                dec         has_xp_sampled phot_bp_mean_mag phot_g_mean_mag phot_rp_mean_mag
                             deg                deg                              mag              mag             mag
    ----------------- ------------------ ------------------ -------------- ---------------- --------------- ----------------
    66727234683960320 56.760485086776846 24.149991010998228           True        14.734505       14.433147        13.954827
    65214031805717376     56.74561610052 24.089174782613686           True        12.338661       11.940813        11.368548
    65225851555715328 56.726951308177455 24.111718134110838           True        14.627676        13.91212        13.091035
    65226195153096192 56.736700233543914 24.149504345515066           True        14.272486       13.613182        12.804853
    >>> len(spectra)
    4
    >>> spectra[66727234683960320]
    wavelength      flux       flux_error
        nm      W / (nm m2)   W / (nm m2)
    ---------- ------------- -------------
         336.0 4.1858373e-17 5.8010027e-18
         338.0  4.101217e-17  4.343636e-18
         340.0  3.499973e-17 3.4906054e-18
         342.0 3.0911544e-17 3.0178371e-18
           ...           ...           ...
        1014.0  1.742315e-17  4.146798e-18
        1016.0 1.5590336e-17  4.358966e-18
        1018.0 1.3888942e-17 4.2265315e-18
        1020.0  1.344579e-17 4.1775913e-18
    """
    # Unlimited rows.
    Gaia.ROW_LIMIT = -1
    # we get the data from GAIA DR3
    Gaia.MAIN_GAIA_TABLE = "gaiadr3.gaia_source"
    try:
        # Query for the catalog to search area with coordinates in FOV of optics.
        job = Gaia.launch_job_async(
            f"SELECT source_id, ra, dec, has_xp_sampled, phot_bp_mean_mag, phot_g_mean_mag, phot_rp_mean_mag \
        FROM gaiadr3.gaia_source \
        WHERE CONTAINS(POINT('ICRS', ra, dec),CIRCLE('ICRS',{right_ascension},{declination},{fov_radius}))=1 \
        AND has_xp_sampled = 'True'"
        )

        # get the results from the query job
        result: Table = job.get_results()

        # set parameters to load data from Gaia catalog
        retrieval_type = "XP_SAMPLED"
        data_release = "Gaia DR3"
        data_structure = "COMBINED"

        # load spectra from stars
        spectra_dct: dict[str, list[tree.Table]] = Gaia.load_data(
            ids=result["source_id"],
            retrieval_type=retrieval_type,
            data_release=data_release,
            data_structure=data_structure,
        )
    except requests.HTTPError as exc:
        raise ConnectionError(
            "Error when trying to communicate with the Gaia database"
        ) from exc
    # save key
    key = f"{retrieval_type}_{data_structure}.xml"

    spectra: dict[int, Table] = {}
    for spectrum in spectra_dct[key]:
        # Get 'table' from the Gaia testbench
        table: Table = spectrum.to_table()

        source_id = int(spectrum.get_field_by_id("source_id").value)
        spectra[source_id] = table

    return result, spectra


def retrieve_from_gaia(
    right_ascension: float,
    declination: float,
    fov_radius: float,
) -> xr.Dataset:
    """Retrieve objects from GAIA Catalog for giver coordinates and FOV.

    Data variable/coordinates description:
    * ``source_id``: Unique source identifier of the source
    * ``ra``: Barycentric right ascension of the source.
    * ``dec``: Barycentric Declination of the source.
    * ``has_xp_sampled``: Flag indicating the availability of mean BP/RP spectrum in sampled form for this source
    * ``phot_bp_mean_mag``: Mean magnitude in the integrated BP band.
    * ``phot_g_mean_mag``: Mean magnitude in the G band.
    * ``phot_rp_mean_mag``: Mean magnitude in the integrated RP band.

    Parameters
    ----------
    right_ascension: float
        RA coordinate in degree.
    declination: float
        DEC coordinate in degree.
    fov_radius: float
        FOV radius of telescope optics.

    Returns
    -------
    Dataset

    Raises
    ------
    ConnectionError
        If the connection to the GAIA database cannot be established.

    Notes
    -----
    More information about the GAIA catalog at these links:
    * https://gea.esac.esa.int/archive/documentation/GDR3/
    * https://gea.esac.esa.int/archive/documentation/GDR3/Gaia_archive/chap_datamodel/sec_dm_main_source_catalogue/ssec_dm_gaia_source.html

    Examples
    --------
    >>> ds = retrieve_from_gaia(
    ...     right_ascension=56.75,
    ...     declination=24.1167,
    ...     fov_radius=0.05,
    ... )
    >>> ds
    <xarray.Dataset>
    Dimensions:           (source_id: 345, wavelength: 343)
    Coordinates:
      * source_id         (source_id) int64 66504343062472704 ... 65296357738731008
      * wavelength        (wavelength) float64 336.0 338.0 ... 1.018e+03 1.02e+03
    Data variables:
        ra                (source_id) float64 57.15 57.18 57.23 ... 56.4 56.42 56.39
        dec               (source_id) float64 23.82 23.83 23.88 ... 24.43 24.46
        has_xp_sampled    (source_id) bool True True True True ... True True True
        phot_bp_mean_mag  (source_id) float32 11.49 14.13 15.22 ... 11.51 8.727
        phot_g_mean_mag   (source_id) float32 10.66 13.8 14.56 ... 14.79 11.09 8.55
        phot_rp_mean_mag  (source_id) float32 9.782 13.29 13.78 ... 14.19 10.5 8.229
        flux              (source_id, wavelength) float32 2.228e-16 ... 3.693e-15
        flux_error        (source_id, wavelength) float32 7.737e-17 ... 2.783e-16
    """
    positions_table: Table
    spectra_dct: dict[int, Table]
    positions_table, spectra_dct = _retrieve_objects_from_gaia(
        right_ascension=right_ascension,
        declination=declination,
        fov_radius=fov_radius,
    )

    # Convert data from Gaia into a dataset
    positions: xr.Dataset = positions_table.to_pandas(index="source_id").to_xarray()
    spectra: xr.Dataset = xr.concat(
        [
            spectrum_table.to_pandas(index="wavelength")
            .to_xarray()
            .assign_coords(source_id=source_id)
            for source_id, spectrum_table in spectra_dct.items()
        ],
        dim="source_id",
    )

    ds: xr.Dataset = xr.merge([positions, spectra])

    # Add units
    first_spectrum: Table = next(iter(spectra_dct.values()))

    ds["wavelength"].attrs = {"units": str(first_spectrum["wavelength"].unit)}
    ds["flux"].attrs = {"units": str(first_spectrum["flux"].unit)}
    ds["flux_error"].attrs = {"units": str(first_spectrum["flux_error"].unit)}
    # ds["flux_photlam"].attrs = {"units": str(first_spectrum["flux_photlam"].unit)}

    ds["ra"].attrs = {
        "name": "Right Ascension",
        "units": str(positions_table["ra"].unit),
    }
    ds["dec"].attrs = {"name": "Declination", "units": str(positions_table["dec"].unit)}
    ds["phot_bp_mean_mag"].attrs = {
        "name": "Mean magnitude in the integrated BP band",
        "units": str(positions_table["phot_bp_mean_mag"].unit),
    }
    ds["phot_g_mean_mag"].attrs = {
        "name": "Mean magnitude in the G band",
        "units": str(positions_table["phot_g_mean_mag"].unit),
    }
    ds["phot_rp_mean_mag"].attrs = {
        "name": "Mean magnitude in the integrated RP band",
        "units": str(positions_table["phot_rp_mean_mag"].unit),
    }

    return ds


def load_objects_from_gaia(
    right_ascension: float,
    declination: float,
    fov_radius: float,
    band: GaiaPassBand = GaiaPassBand.BluePhotometer,
) -> xr.Dataset:
    """Load objects from GAIA Catalog for given coordinates and FOV.

    Parameters
    ----------
    right_ascension : float
        RA coordinate in degree.
    declination : float
        DEC coordinate in degree.
    fov_radius : float
        FOV radius of telescope optics.
    band : GaiaPassBand
        Define the passband to select.

    Returns
    -------
    Dataset
        Dataset object in the FOV at given coordinates found by the GAIA catalog.

    Examples
    --------
    >>> ds = load_objects_from_gaia(
    ...     right_ascension=56.75,
    ...     declination=24.1167,
    ...     fov_radius=0.5,
    ... )
    >>> ds
    <xarray.Dataset>
    Dimensions:     (ref: 345, wavelength: 343)
    Coordinates:
      * ref         (ref) int64 0 1 2 3 4 5 6 7 ... 337 338 339 340 341 342 343 344
      * wavelength  (wavelength) float64 336.0 338.0 340.0 ... 1.018e+03 1.02e+03
    Data variables:
        x           (ref) float64 1.334e+03 1.434e+03 ... -1.271e+03 -1.381e+03
        y           (ref) float64 -1.009e+03 -956.1 -797.1 ... 1.195e+03 1.309e+03
        weight      (ref) float64 11.49 14.13 15.22 14.56 ... 15.21 11.51 8.727
        flux        (ref, wavelength) float64 2.228e-16 2.432e-16 ... 3.693e-15
    """
    # Get all sources and spectrum in one Dataset
    ds_from_gaia: xr.Dataset = retrieve_from_gaia(
        right_ascension=right_ascension,
        declination=declination,
        fov_radius=fov_radius,
    )

    # Convert 'flux' from W / (nm m2) to ph / (s AA cm2)
    flux_converted: xr.DataArray = convert_spectrum(flux=ds_from_gaia["flux"])

    # Convert the positions to arcseconds and center around 0.0, 0.0.
    # The centering is necessary because ScopeSIM cannot actually 'point' the telescope.
    ra_arcsec: u.Quantity = u.Quantity(ds_from_gaia["ra"], unit=u.deg).to(u.arcsec)
    dec_arcsec: u.Quantity = u.Quantity(ds_from_gaia["dec"], unit=u.deg).to(u.arcsec)

    x: u.Quantity = ra_arcsec  # - ra_arcsec.mean()
    y: u.Quantity = dec_arcsec  # - dec_arcsec.mean()

    # Get weights
    weights_from_gaia: xr.DataArray = ds_from_gaia[band.get_magnitude_key()]

    num_sources = len(ds_from_gaia["source_id"])
    ref_sequence: Sequence[int] = range(num_sources)

    ds = xr.Dataset(coords={"ref": ref_sequence})
    ds["x"] = xr.DataArray(np.asarray(x), dims="ref", attrs={"units": str(x.unit)})
    ds["y"] = xr.DataArray(np.asarray(y), dims="ref", attrs={"units": str(y.unit)})
    ds["weight"] = xr.DataArray(
        np.asarray(weights_from_gaia, dtype=float),
        dims="ref",
        attrs={"units": weights_from_gaia.attrs["units"]},
    )
    ds["flux"] = xr.DataArray(
        np.asarray(flux_converted, dtype=float),
        dims=["ref", "wavelength"],
        coords={"wavelength": flux_converted["wavelength"]},
        attrs={"units": flux_converted.attrs["units"]},
    )

    return ds


# TODO: add information about magnitude
# TODO: add option to select filter to compute apparent magnitude
# TODO: add option to select different catalogue versions


def load_star_map(
    detector: Detector,
    right_ascension: float,
    declination: float,
    fov_radius: float,
    band: Literal["blue_photometer", "gaia_band", "red_photometer"] = "blue_photometer",
):
    """Generate scene from scopesim Source object loading stars from the GAIA catalog.

    Parameters
    ----------
    detector : Detector
        Pyxel Detector object.
    right_ascension : float
        RA coordinate in degree.
    declination : float
        DEC coordinate in degree.
    fov_radius : float
        FOV radius of telescope optics.
    band : 'blue_photometer', 'gaia_band' or 'red_photometer'
        Define the band to use.
        * 'blue_photometer' is the band from 330 nm to 680 nm
        * 'gaia_band' is the band from 330 nm to 1050 nm
        * 'red_photometer' is the band from 640 nm to 1050 nm
    """
    band_pass: GaiaPassBand = GaiaPassBand(band)

    ds: xr.Dataset = load_objects_from_gaia(
        right_ascension=right_ascension,
        declination=declination,
        fov_radius=fov_radius,
        band=band_pass,
    )

    detector.scene.add_source(ds)



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.


import numpy as np
import pytest
import statsmodels.api as sm
import xarray as xr

from pyxel.detectors import CCD, CCDGeometry, Characteristics, Environment
from pyxel.models.data_processing import linear_regression


@pytest.fixture
def ccd_10x10() -> CCD:
    """Create a valid CCD detector."""
    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector.set_readout(times=[1.0, 2.0, 3.0])

    return detector


def test_raise_error():
    """Test that error is raised when less than two times are given."""

    detector = CCD(
        geometry=CCDGeometry(
            row=10,
            col=10,
            total_thickness=40.0,
            pixel_vert_size=10.0,
            pixel_horz_size=10.0,
        ),
        environment=Environment(),
        characteristics=Characteristics(),
    )
    detector.set_readout(times=[1.0])

    with pytest.raises(ValueError, match="Expecting at least 2 steps"):
        linear_regression(detector=detector)


@pytest.fixture
def image_3d() -> xr.DataArray:
    """Create fake image for three different times."""

    rng = np.random.default_rng(seed=100)
    relative_time = [1.0, 2.0, 3.0]
    image = xr.DataArray(
        rng.random((3, 10, 10)),
        dims=["time", "y", "x"],
        coords={"time": relative_time, "y": range(10), "x": range(10)},
    )
    return image


@pytest.fixture
def expected_result(image_3d: xr.DataArray) -> xr.Dataset:
    """Compute expected result with 'statsmodels'."""

    def compute_linear_regression(
        y: np.ndarray, x: np.ndarray
    ) -> tuple[float, float, float, float, float]:
        model = sm.OLS(endog=y, exog=sm.add_constant(x))
        results = model.fit()

        intercept, slope = results.params
        r2 = results.rsquared
        intercept_std, slope_std = results.bse

        return intercept, slope, r2, intercept_std, slope_std

    all_results = xr.apply_ufunc(
        compute_linear_regression,
        image_3d,
        image_3d["time"],
        input_core_dims=[["time"], ["time"]],
        output_core_dims=[[], [], [], [], []],
        vectorize=True,
    )
    intercept, slope, r2, intercept_std, slope_std = all_results

    ds = xr.Dataset()
    ds["intercept"] = intercept
    ds["slope"] = slope
    ds["r2"] = r2
    ds["intercept_std"] = intercept_std
    ds["slope_std"] = slope_std

    return ds


def test_linear_regression(
    ccd_10x10: CCD,
    image_3d: xr.DataArray,
    expected_result: xr.Dataset,
):
    """Test model 'linear_regression'."""

    detector = ccd_10x10
    steps = detector.readout_properties.steps
    absolute_time = detector.readout_properties.times

    for i, (time, step) in enumerate(zip(absolute_time, steps)):
        detector.time = time
        detector.time_step = step
        detector.pipeline_count = i

        detector.image.array = image_3d.isel(time=i).to_numpy()
        linear_regression(detector=detector, data_structure="image")

    dataset = detector.data["/linear_regression/image"].to_dataset()

    # check that both datasets are the same
    xr.testing.assert_allclose(dataset, expected_result)
